import numpy as np
from warnings import warn
from typing import List, Callable
from desdeo_emo.selection.SelectionBase import SelectionBase
from desdeo_emo.population.Population import Population
from desdeo_emo.utilities.ReferenceVectors import ReferenceVectors


class IOPIS_APD_Select(SelectionBase):
    """The selection operator for the IOPIS/RVEA algorithm.

    Parameters
    ----------
    pop : Population
        The population instance
    time_penalty_function : Callable
        A function that returns the time component in the penalty function.
    alpha : float, optional
        The RVEA alpha parameter, by default 2
    """

    def __init__(
        self,
        time_penalty_function: Callable,
        scalarization_methods: List,
        alpha: float = 2,
    ):
        self.time_penalty_function = time_penalty_function
        self.alpha = 2
        self.n_of_objectives = len(scalarization_methods)
        self.scalarization_methods = scalarization_methods

    def do(
        self, pop: Population, vectors: ReferenceVectors, reference_point: np.ndarray
    ) -> List[int]:
        """Select individuals for mating on basis of Angle penalized distance.

        Parameters
        ----------
        pop : Population
            The current population.
        vectors : ReferenceVectors
            Class instance containing reference vectors.

        Returns
        -------
        List[int]
            List of indices of the selected individuals
        """
        partial_penalty_factor = self._partial_penalty_factor()
        refV = vectors.neighbouring_angles_current
        fitness = np.asarray(
            [
                scalar(pop.fitness, reference_point)
                for scalar in self.scalarization_methods
            ]
        ).T
        fmin = np.amin(fitness, axis=0)
        translated_fitness = fitness - fmin
        fitness_norm = np.linalg.norm(translated_fitness, axis=1)
        # TODO check if you need the next line
        fitness_norm = np.repeat(fitness_norm, len(translated_fitness[0, :])).reshape(
            len(fitness), len(fitness[0, :])
        )
        # Convert zeros to eps to avoid divide by zero.
        # Has to be checked!
        fitness_norm[fitness_norm == 0] = np.finfo(float).eps
        normalized_fitness = np.divide(
            translated_fitness, fitness_norm
        )  # Checked, works.
        cosine = np.dot(normalized_fitness, np.transpose(vectors.values))
        if cosine[np.where(cosine > 1)].size:
            warn("RVEA.py line 60 cosine larger than 1 decreased to 1")
            cosine[np.where(cosine > 1)] = 1
        if cosine[np.where(cosine < 0)].size:
            warn("RVEA.py line 64 cosine smaller than 0 increased to 0")
            cosine[np.where(cosine < 0)] = 0
        # Calculation of angles between reference vectors and solutions
        theta = np.arccos(cosine)
        # Reference vector asub_population_indexssignment
        assigned_vectors = np.argmax(cosine, axis=1)
        selection = np.array([], dtype=int)
        # Selection
        # Convert zeros to eps to avoid divide by zero.
        # Has to be checked!
        refV[refV == 0] = np.finfo(float).eps
        for i in range(0, len(vectors.values)):
            sub_population_index = np.atleast_1d(
                np.squeeze(np.where(assigned_vectors == i))
            )
            sub_population_fitness = translated_fitness[sub_population_index]
            if len(sub_population_fitness > 0):
                # APD Calculation
                angles = theta[sub_population_index, i]
                angles = np.divide(angles, refV[i])  # This is correct.
                # You have done this calculation before. Check with fitness_norm
                # Remove this horrible line
                sub_pop_fitness_magnitude = np.sqrt(
                    np.sum(np.power(sub_population_fitness, 2), axis=1)
                )
                apd = np.multiply(
                    np.transpose(sub_pop_fitness_magnitude),
                    (1 + np.dot(partial_penalty_factor, angles)),
                )
                minidx = np.where(apd == np.nanmin(apd))
                if np.isnan(apd).all():
                    continue
                selx = sub_population_index[minidx]
                if selection.shape[0] == 0:
                    selection = np.hstack((selection, np.transpose(selx[0])))
                else:
                    selection = np.vstack((selection, np.transpose(selx[0])))
        return selection.squeeze()

    def _partial_penalty_factor(self) -> float:
        """Calculate and return the partial penalty factor for APD calculation.
            This calculation does not include the angle related terms, hence the name.
            If the calculated penalty is outside [0, 1], it will round it up/down to 0/1

        Returns
        -------
        float
            The partial penalty value
        """
        penalty = ((self.time_penalty_function()) ** self.alpha) * self.n_of_objectives
        if penalty < 0:
            penalty = 0
        if penalty > 1:
            penalty = 1
        return penalty



"""This module provides implementations of various crossover and mutation operators.
"""

__all__ = [
    "SBX_xover",
    "BP_mutation",
    "EvoNNRecombination",
    "EvoDN2Recombination",
    "BioGP_xover",
    "BioGP_mutation",
    "SinglePoint_Xover",
    "SinglePoint_Mutation",
]

from desdeo_emo.recombination.biogp_mutation import BioGP_mutation
from desdeo_emo.recombination.biogp_xover import BioGP_xover
from desdeo_emo.recombination.BoundedPolynomialMutation import BP_mutation
from desdeo_emo.recombination.evodn2_xover_mutation import EvoDN2Recombination
from desdeo_emo.recombination.evonn_xover_mutation import EvoNNRecombination
from desdeo_emo.recombination.SimulatedBinaryCrossover import SBX_xover
from desdeo_emo.recombination.SinglePointCrossoverMutation import SinglePoint_Xover, SinglePoint_Mutation



from random import choice, random
from math import ceil
from typing import Callable, Dict, Type, Tuple

import numpy as np
from desdeo_problem.surrogatemodels.SurrogateModels import BaseRegressor, ModelError
from sklearn.metrics import mean_squared_error, mean_squared_log_error, r2_score

from desdeo_emo.EAs.BaseEA import BaseEA
from desdeo_emo.EAs.PPGA import PPGA
from desdeo_emo.EAs.TournamentEA import TournamentEA

# from graphviz import Digraph, Source
import pandas as pd
from desdeo_emo.surrogatemodels.Problem import surrogateProblem
from desdeo_emo.population.SurrogatePopulation import SurrogatePopulation
from desdeo_emo.recombination.biogp_xover import BioGP_xover
from desdeo_emo.recombination.biogp_mutation import BioGP_mutation
from desdeo_emo.utilities.plotlyanimate import animate_init_, animate_next_


def negative_r2_score(y_true, y_pred):
    return -r2_score(y_true, y_pred)


class Node:
    """A node object representing a function or terminal node in the tree.

    Parameters
    ----------
    value : function, str or float
        A function node has as its value a function. Terminal nodes contain variables
        which are either float or str.
    depth : int
        The depth the node is at.
    function_set : array_like
        The function set to use when creating the trees.
    terminal_set : array_like
        The terminals (variables and constants) to use when creating the trees.

    """

    def __init__(
        self,
        value,
        depth,
        max_depth,
        max_subtrees,
        prob_terminal,
        function_set,
        terminal_set,
    ):
        self.value = value
        self.depth = depth
        self.max_depth = max_depth
        self.max_subtrees = max_subtrees
        self.prob_terminal = prob_terminal
        self.function_set = function_set
        self.terminal_set = terminal_set
        self.nodes = []
        self.nodes_at_depth = None
        self.total_depth = None
        self.roots = []

    def predict(self, decision_variables=None):

        if callable(self.value):
            values = [root.predict(decision_variables) for root in self.roots]
            if self.depth == 0:
                self.value = sum(values)
                return self.value
            else:
                return self.value(*values)

        else:
            if isinstance(decision_variables, np.ndarray) and isinstance(
                self.value, str
            ):
                return decision_variables[
                    :, int("".join(filter(str.isdigit, self.value))) - 1
                ]
            if isinstance(self.value, str):
                return np.asarray(decision_variables[self.value]).reshape(-1, 1)
            else:
                return np.full((decision_variables.shape[0], 1), self.value)

    def node_label(self):  # return string label
        if callable(self.value):
            return self.value.__name__
        else:
            return str(self.value)

    def draw(self, dot, count):  # dot & count are lists in order to pass "by reference"
        node_name = str(count[0])
        dot[0].node(node_name, self.node_label())

        for node in self.roots:
            count[0] += 1
            dot[0].edge(node_name, str(count[0]))
            node.draw(dot, count)

    """def draw_tree(self, name="tree", footer=""):
        dot = [Digraph()]
        dot[0].attr(kw="graph", label=footer)
        count = [0]
        self.draw(dot, count)
        Source(dot[0], filename=name + ".gv", format="png").render()"""

    def get_sub_nodes(self):
        """Get all nodes belonging to the subtree under the current node.

        Returns
        -------
        nodes : list
            A list of nodes in the subtree.

        """
        nodes = []
        nodes_at_depth = {}
        stack = [self]
        while stack:
            cur_node = stack[0]
            depth = cur_node.depth
            if cur_node.depth not in nodes_at_depth:
                nodes_at_depth[cur_node.depth] = []
            nodes_at_depth[cur_node.depth].append(cur_node)
            stack = stack[1:]
            nodes.append(cur_node)
            if len(cur_node.roots) > 0:
                depth += 1
            for child in cur_node.roots:
                child.depth = depth
                stack.append(child)

        self.nodes_at_depth = nodes_at_depth
        self.total_depth = max(key for key in self.nodes_at_depth.keys())

        return nodes

    def grow_tree(self, max_depth=None, method="grow", depth=0, ind=None):
        """Create a random tree recursively using either grow or full method.

        Parameters
        ----------
        max_depth : int
            The maximum depth of the tree.
        method : str
            Methods: 'grow', 'full'.
            For the 'grow' method, nodes are chosen at random from both functions and
            terminals.
            The 'full' method chooses nodes from the function set until the max depth
            is reached,
            and then terminals are chosen.
        depth : int
            Current depth.
        ind : :obj:
            The starting node from which to begin growing trees.

        """
        node = None
        if max_depth is None:
            max_depth = self.max_depth
        if depth == 0:
            if ind is None:
                ind = LinearNode(value="linear")
            num_subtrees = self.max_subtrees
            for i in range(len(ind.roots), num_subtrees):
                node = self.grow_tree(max_depth, method, depth=depth + 1)
                ind.roots.append(node)

        # Make terminal node
        elif depth >= max_depth or method == "grow" and random() < self.prob_terminal:
            node = Node(
                depth=depth,
                function_set=self.function_set,
                terminal_set=self.terminal_set,
                value=None,
                max_depth=self.max_depth,
                max_subtrees=self.max_subtrees,
                prob_terminal=self.prob_terminal,
            )
            node.value = choice(node.terminal_set)

        # Make function node
        else:
            node = Node(
                depth=depth,
                function_set=self.function_set,
                terminal_set=self.terminal_set,
                value=None,
                max_depth=self.max_depth,
                max_subtrees=self.max_subtrees,
                prob_terminal=self.prob_terminal,
            )
            node.value = choice(node.function_set)

            for i in range(node.value.__code__.co_argcount):  # Check arity
                root = self.grow_tree(max_depth, method, depth=depth + 1)
                node.roots.append(root)

        return node


class LinearNode(Node):
    """The parent node of the tree, from which a number of subtrees emerge, as defined
    by the user. The linear node takes a weighted sum of the output from the subtrees
    and
    also uses a bias value. The weights and the bias are calculated by the linear least
    square technique.

    Parameters
    ----------
    value : function, str or float
        A function node has as its value a function. Terminal nodes contain variables
        which are either float or str.
    depth : int
        The depth the node is at.
    """

    def __init__(
        self,
        max_depth,
        max_subtrees,
        prob_terminal,
        function_set,
        terminal_set,
        error_lim,
        complexity_scalar=0.5,
        value="linear",
        depth=0,
    ):
        super().__init__(
            value=value,
            depth=depth,
            max_depth=max_depth,
            max_subtrees=max_subtrees,
            prob_terminal=prob_terminal,
            function_set=function_set,
            terminal_set=terminal_set,
        )
        self.nodes_at_depth = {}
        self.error_lim = error_lim
        self.complexity_scalar = complexity_scalar
        self.out = None
        self.complexity = None
        self.fitness = None
        self.linear = None

    def calculate_linear(self, X_train, y_train):

        sub_trees = []
        # Stack outputs of subtrees to form weight matrix
        for root in self.roots:
            sub_trees.append(root.predict(X_train))

        sub_trees = np.hstack(sub_trees)
        axis = None
        if sub_trees.ndim > 1:
            axis = 1
        sub_trees = np.insert(sub_trees, 0, 1, axis=axis)

        weights, *_ = np.linalg.lstsq(sub_trees, y_train, rcond=None)
        self.linear = weights
        out = np.dot(sub_trees, weights)

        # Error reduction ratio
        q, r = np.linalg.qr(sub_trees)
        s = np.linalg.lstsq(q, y_train, rcond=None)[0]
        error = np.divide((s ** 2 * np.sum(q * q, axis=0)), np.sum(out * out, axis=0))

        # If error reduction ration < err_lim, delete root and grow new one
        for i, err in enumerate(error[1:]):
            if err < self.error_lim:
                del self.roots[i]
                self.grow_tree(max_depth=self.max_depth, method="grow", ind=self)

        self.nodes = self.get_sub_nodes()

        num_func_nodes = sum(1 for node in self.nodes if callable(node.value))

        complexity = (
            self.complexity_scalar * self.total_depth
            + (1 - self.complexity_scalar) * num_func_nodes
        )

        self.out = out
        self.complexity = complexity

        return [self.out, self.complexity]


class BioGP(BaseRegressor):
    def __init__(
        self,
        training_algorithm: Type[BaseEA] = PPGA,
        pop_size: int = 500,
        probability_crossover: float = 0.9,
        probability_mutation: float = 0.3,
        max_depth: int = 5,
        max_subtrees: int = 4,
        prob_terminal: float = 0.5,
        complexity_scalar: float = 0.5,
        error_lim: float = 0.001,
        init_method: str = "ramped_half_and_half",
        model_selection_criterion: str = "min_error",
        loss_function: str = "mse",
        single_obj_generations: int = 10,
        function_set=("add", "sub", "mul", "div"),
        terminal_set=None,
    ):
        loss_functions = {
            "mse": mean_squared_error,
            "msle": mean_squared_log_error,
            "neg_r2": negative_r2_score,
        }
        self.function_map = {
            "add": self.add,
            "sub": self.sub,
            "mul": self.mul,
            "div": self.div,
            "sqrt": self.sqrt,
            "log": self.log,
            "sin": self.sin,
            "cos": self.cos,
            "tan": self.tan,
            "neg": self.neg,
        }
        self.training_algorithm = training_algorithm
        self.pop_size: int = pop_size
        self.probability_crossover: float = probability_crossover
        self.probability_mutation: float = probability_mutation
        self.max_depth: int = max_depth
        self.max_subtrees: int = max_subtrees
        self.prob_terminal: float = prob_terminal
        self.complexity_scalar: float = complexity_scalar
        self.error_lim: float = error_lim
        self.init_method: str = init_method
        self.model_selection_criterion: str = model_selection_criterion
        self.loss_function_str: str = loss_function
        self.loss_function: Callable = loss_functions[loss_function]
        self.single_obj_generations: int = single_obj_generations
        self.function_set_str: Tuple[str] = function_set
        self.function_set = []
        for function in function_set:
            self.function_set.append(self.function_map[function])
        self.terminal_set: Tuple = terminal_set
        #
        self.X: np.ndarray = None
        self.y: np.ndarray = None
        self.model_trained: bool = False
        # Model Parameters
        self.tree: LinearNode = None
        self.performance: Dict = {"RMSE": None, "R^2": None}
        self.model_population = None

    def fit(self, X: pd.DataFrame, y: pd.DataFrame):
        if X.shape[0] != y.shape[0]:
            msg = (
                f"Ensure that the number of samples in X and y are the same"
                f"Number of samples in X = {X.shape[0]}"
                f"Number of samples in y = {y.shape[0]}"
            )
            raise ModelError(msg)
        self.X = X
        self.y = y
        if self.terminal_set is None:
            self.terminal_set = X.columns.tolist()

        # Create problem
        problem = surrogateProblem(performance_evaluator=self._model_performance)
        problem.n_of_objectives = 2
        # Create Population
        initial_pop = self._create_individuals()
        population = SurrogatePopulation(
            problem, self.pop_size, initial_pop, None, None, None
        )
        population.xover = BioGP_xover(probability_crossover=self.probability_crossover)
        population.mutation = BioGP_mutation(
            probability_mutation=self.probability_mutation
        )
        # Do single objective evolution
        tournament_evolver = TournamentEA(
            problem,
            initial_population=population,
            population_size=self.pop_size,
            n_gen_per_iter=self.single_obj_generations,
            n_iterations=1,
        )
        figure = animate_init_(
            tournament_evolver.population.objectives, filename="BioGP.html"
        )
        while tournament_evolver.continue_evolution():
            tournament_evolver.iterate()
            figure = animate_next_(
                tournament_evolver.population.objectives,
                figure,
                filename="BioGP.html",
                generation=tournament_evolver._iteration_counter,
            )
        population = tournament_evolver.population

        # Do bi-objective evolution
        evolver = self.training_algorithm(
            problem,
            initial_population=population,
            population_size=self.pop_size,
            n_gen_per_iter=10,
            n_iterations=10,
        )
        while evolver.continue_evolution():
            evolver.iterate()
            figure = animate_next_(
                evolver.population.objectives,
                figure,
                filename="BioGP.html",
                generation=evolver._iteration_counter
                + tournament_evolver._iteration_counter,
            )
        self.model_population = evolver.population
        # Selection
        self.select()
        self.model_trained = True

    def _create_individuals(self):
        individuals = []
        if self.init_method == "ramped_half_and_half":
            for md in range(ceil(self.max_depth / 2), self.max_depth + 1):
                for i in range(int(self.pop_size / (self.max_depth + 1))):

                    ind = LinearNode(
                        value="linear",
                        max_depth=self.max_depth,
                        max_subtrees=self.max_subtrees,
                        prob_terminal=self.prob_terminal,
                        function_set=self.function_set,
                        terminal_set=self.terminal_set,
                        error_lim=self.error_lim,
                    )
                    ind.grow_tree(max_depth=md, method="grow", ind=ind)
                    individuals.append(ind)

                for i in range(int(self.pop_size / (self.max_depth + 1))):

                    ind = LinearNode(
                        value="linear",
                        max_depth=self.max_depth,
                        max_subtrees=self.max_subtrees,
                        prob_terminal=self.prob_terminal,
                        function_set=self.function_set,
                        terminal_set=self.terminal_set,
                        error_lim=self.error_lim,
                    )
                    ind.grow_tree(max_depth=md, method="full", ind=ind)
                    individuals.append(ind)

        elif self.init_method == "full":
            for i in range(self.pop_size):
                ind = LinearNode(
                    value="linear",
                    max_depth=self.max_depth,
                    max_subtrees=self.max_subtrees,
                    prob_terminal=self.prob_terminal,
                    function_set=self.function_set,
                    terminal_set=self.terminal_set,
                    error_lim=self.error_lim,
                )
                ind.grow_tree(method="full", ind=ind)
                individuals.append(ind)

        elif self.init_method == "grow":
            for i in range(self.pop_size):
                ind = LinearNode(
                    value="linear",
                    max_depth=self.max_depth,
                    max_subtrees=self.max_subtrees,
                    prob_terminal=self.prob_terminal,
                    function_set=self.function_set,
                    terminal_set=self.terminal_set,
                    error_lim=self.error_lim,
                )
                ind.grow_tree(method="grow", ind=ind)
                individuals.append(ind)

        return np.asarray(individuals).reshape(-1, 1)

    def _model_performance(
        self, trees: LinearNode, X: np.ndarray = None, y: np.ndarray = None
    ):
        if trees is None and self.model_trained is False:
            msg = "Model has not been trained yet"
            raise ModelError(msg)
        if trees is None:
            trees = self.tree
        if X is None:
            X = self.X
            y = self.y
        if len(trees) > 1:
            loss = []
            complexity = []
            for tree in trees:
                y_pred, ind_complexity = tree[0].calculate_linear(X, y)
                loss.append(self.loss_function(y, y_pred))
                complexity.append(ind_complexity)
        else:
            y_pred, complexity = trees[0].calculate_linear(X, y)
            loss = self.loss_function(y, y_pred)
        return np.asarray((loss, complexity)).T

    def predict(self, X: np.ndarray):
        if isinstance(X, (np.ndarray, list)):
            X = pd.DataFrame(X, columns=self.X.columns)
        sub_trees = []
        # Stack outputs of subtrees to form weight matrix
        for root in self.tree.roots:
            sub_tree = root.predict(X)
            if sub_tree.size == 1:
                sub_tree = sub_tree.reshape(1)
            sub_trees.append(sub_tree)

        sub_trees = np.hstack(sub_trees)
        axis = None
        if sub_trees.ndim > 1:
            axis = 1
        sub_trees = np.insert(sub_trees, 0, 1, axis=axis)

        y = np.dot(sub_trees, self.tree.linear)

        if isinstance(y, float):
            y = np.asarray([y])

        return y, np.zeros_like(y)

    def select(self):
        if self.model_selection_criterion == "min_error":
            # Return the model with the lowest error
            selected = np.argmin(self.model_population.objectives[:, 0])
            # print(self.model_population.objectives)
        else:
            raise ModelError("Selection criterion not recognized. Use 'min_error'.")
        self.tree = self.model_population.individuals[selected][0]
        y_pred = self.predict(X=self.X)[0]
        self.performance["RMSE"] = np.sqrt(mean_squared_error(self.y, y_pred))
        self.performance["R^2"] = r2_score(self.y, y_pred)

    @staticmethod
    def add(x, y):
        return np.add(x, y)

    @staticmethod
    def sub(x, y):
        return np.subtract(x, y)

    @staticmethod
    def mul(x, y):
        return np.multiply(x, y)

    @staticmethod
    def div(x, y):
        y[y == 0] = 1.0
        return np.divide(x, y)

    @staticmethod
    def sqrt(x):
        return np.sqrt(np.abs(x))

    @staticmethod
    def log(x):
        return np.where(np.abs(x) > 0.001, np.log(np.abs(x)), 0.0)

    @staticmethod
    def sin(x):
        return np.sin(x)

    @staticmethod
    def cos(x):
        return np.cos(x)

    @staticmethod
    def tan(x):
        return np.tan(x)

    @staticmethod
    def neg(x):
        return np.negative(x)



"""This module provides implementations of EAs which can be used for training
surrogate models.
"""

__all__ = ["BioGP", "EvoNN", "EvoDN2", "surrogateProblem"]

from desdeo_emo.surrogatemodels.BioGP import BioGP
from desdeo_emo.surrogatemodels.EvoNN import EvoNN
from desdeo_emo.surrogatemodels.EvoDN2 import EvoDN2
from desdeo_emo.surrogatemodels.Problem import surrogateProblem



from typing import Dict, List, Union

import numpy as np
import pandas as pd
from desdeo_emo.population.Population import Population
from desdeo_problem import (
    MOProblem,
    ScalarConstraint,
    ScalarObjective,
    Variable,
    VectorObjective,
)
from desdeo_tools.maps.preference_incorporated_space_RP import (
    IOPIS,
    MultiDMPIS,
    classificationPIS,
)
from desdeo_tools.scalarization import AUG_GUESS_GLIDE, AUG_STOM_GLIDE


class IOPISProblem(MOProblem):
    """A problem class for the IOPIS formulation for interactive optimization.

    This variant uses the classification kind of preference information for
    the creation of the Preference incorporated space (PIS).

    Arguments:
        objectives (List[Union[ScalarObjective, VectorObjective]]): A list containing
            the objectives of the problem.
        variables (List[Variable]): A list containing the variables of the problem.
        nadir (np.ndarray): Nadir point of the problem.
        ideal (np.ndarray): Ideal point of the problem.
        PIS: An instantiated classificationPIS class from desdeo-tools.
        constraints (List[ScalarConstraint], optional): A list of the constraints of the problem. Defaults to None.
    """

    def __init__(
        self,
        objectives: List[Union[ScalarObjective, VectorObjective]],
        variables: List[Variable],
        nadir: np.ndarray,
        ideal: np.ndarray,
        PIS_type: str,
        initial_preference: Union[List, Dict],
        constraints: List[ScalarConstraint] = None,
    ):
        super().__init__(
            objectives=objectives,
            variables=variables,
            constraints=constraints,
            nadir=nadir,
            ideal=ideal,
        )
        allowed_PIS_types = ["MultiDM", "Classification", "IOPIS"]
        self.PIS_type = PIS_type
        if PIS_type not in allowed_PIS_types:
            raise ValueError(f"PIS_type should be one of {allowed_PIS_types}")
        if PIS_type == "IOPIS":
            self.PIS = IOPIS(
                utopian=self.ideal * self._max_multiplier,
                nadir=self.nadir * self._max_multiplier,
                scalarizers=[AUG_STOM_GLIDE, AUG_GUESS_GLIDE],
            )
            self.fitness_names = ["STOM Value", "GUESS Value"]
            self.allowable_interaction_types = {
                "Reference point": (
                    "Specify a reference point worse than the utopian point. ",
                    "The STOM and GUESS scalarization functions are used to find "
                    "solutions that approximate the given reference point. "
                    "This results in solutions that represent the trade-offs between "
                    "the optimal solution for STOM and GUESS."
                    " The closer the reference point is to the Pareto front, the smaller "
                    "the spread of the solutions is.",
                )
            }
        elif PIS_type == "Classification":
            self.PIS = classificationPIS(
                utopian=self.ideal * self._max_multiplier,
                nadir=self.nadir * self._max_multiplier,
                scalarizers=[AUG_STOM_GLIDE],
            )
            self.fitness_names = ["NIMBUS Value", "STOM Value"]
            self.allowable_interaction_types = {
                "Classification": (
                    "Classify the objectives into 'make better', 'keep same', "
                    "'allow to worsen'."
                )
            }
        elif PIS_type == "MultiDM":
            self.PIS = MultiDMPIS(
                utopian=self.ideal * self._max_multiplier,
                nadir=self.nadir * self._max_multiplier,
                num_DM=2,
            )
            self.fitness_names = [
                "First preference fitness",
                "Second preference fitness",
            ]
            self.allowable_interaction_types = {
                "Reference points": (
                    "Specify two reference points worse than the utopian point. ",
                    "The STOM scalarization functions are used to find "
                    "solutions that approximate the given reference points. "
                    "This results in solutions that represent the trade-offs between "
                    "the optimal solution for the two reference points."
                    " The closer the reference points are to each other, the smaller "
                    "the spread of the solutions is.",
                )
            }
        self.PIS.update_preference(initial_preference)
        self.ideal_fitness = self.PIS(self.ideal * self._max_multiplier)
        self.nadir_fitness = self.PIS(self.nadir * self._max_multiplier)

        self.num_dim_fitness = self.PIS.num_scalarizers

    @property
    def n_of_fitnesses(self) -> int:
        return self.num_dim_fitness

    def evaluate_fitness(self, objective_vectors: np.ndarray) -> np.ndarray:
        """Evaluate objective fitness.

        Arguments:
            objective_vectors (np.ndarray): objective vectors

        Returns:
            np.ndarray: Objective fitness
        """
        return self.PIS(objective_vectors * self._max_multiplier)

    def reevaluate_fitness(self, objective_vectors: np.ndarray) -> np.ndarray:
        """Re-evaluate objective fitness.

        Calls update_ideal with objective_vectors.

        Arguments:
            objective_vectors (np.ndarray): objective vectors

        Returns:
            np.ndarray: Objective fitness
        """
        fitness = self.PIS(objective_vectors * self._max_multiplier)
        self.ideal_fitness = self.PIS(self.ideal * self._max_multiplier)
        self.update_ideal(objective_vectors, fitness)
        return fitness

    def update_preference(self, preference: Union[Dict, List]):
        """Update PIS preference

        Arguments:
            preference (Dict): PIS preferences

        """
        self.PIS.update_preference(preference)

    def update_ideal(self, objective_vectors: np.ndarray, fitness: np.ndarray):
        """Update ideal vector.

        Arguments:
            objective_vectors (np.ndarray): Objective vectors
            fitness (np.ndarray): Fitness values for objective vectors

        """
        self.ideal_fitness = np.amin(np.vstack((self.ideal_fitness, fitness)), axis=0)

        self.ideal = (
            np.amin(
                np.vstack((self.ideal, objective_vectors)) * self._max_multiplier,
                axis=0,
            )
            * self._max_multiplier
        )
        self.PIS.update_map(
            utopian=self.ideal * self._max_multiplier,
            nadir=self.nadir * self._max_multiplier,
        )

    def request_preferences(self, pop):
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=pop.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = pop.problem._max_multiplier
        dimensions_data.loc["ideal"] = pop.ideal_objective_vector
        dimensions_data.loc["nadir"] = pop.nadir_objective_vector
        message = (
            f"Provide the following preference type: \n"
            f"{self.allowable_interaction_types}"
        )
        return dimensions_data, message

    def manage_preferences(self, population: Population, preference=None):
        if self.PIS_type == "IOPIS":
            if not isinstance(preference, dict):
                raise TypeError("Preference type must be dict")
            self.PIS.update_map(
                utopian=population.ideal_objective_vector
                * population.problem._max_multiplier,
                nadir=population.nadir_objective_vector
                * population.problem._max_multiplier,
            )
            self.PIS.update_preference(preference)
            return
        elif self.PIS_type == "MultiDM":
            if not isinstance(preference, list):
                raise TypeError("Preference type must be list")

            self.PIS.update_map(
                utopian=population.ideal_objective_vector
                * population.problem._max_multiplier,
                nadir=population.nadir_objective_vector
                * population.problem._max_multiplier,
            )
            self.PIS.update_preference(preference)
            return
        else:
            raise TypeError("Classification not implemented in this version yet.")



import numpy as np
from typing import List
from desdeo_emo.selection.SelectionBase import SelectionBase
from desdeo_emo.population.Population import Population


class TournamentSelection(SelectionBase):
    """
        Tournament selection operator.

        Parameters
        ----------
        pop : Population
            The population of individuals
        tournament_size : int
            Size of the tournament.
    """
    def __init__(self, pop, tournament_size):
        # initialize
        #self.fitness = pop.fitness
        #self.pop_size = pop.pop_size
        self.tournament_size = tournament_size


    # TODO: add the opt fitness, which is the custom set of fitness values e.g self.local_fitnesss
    def do(self, pop, fitness) -> List[int]:
        """ Performs tournament selections and returns the parents.
        Parameters
        ----------
        pop : Population
            The current population.

        Returns
        -------
        List[int]
            List of indices of the selected individuals
        """
        self.fitness = fitness
        self.pop_size = self.fitness.shape[0]
        parents = []
        for i in range(int(self.pop_size)): 
            parents.append(
                np.asarray(
                    self._tour_select(), 
                    self._tour_select(),
            ))
        return parents


    def _tour_select(self):
        """Tournament selection. Choose number of individuals to participate
        and select the one with the best fitness.

        Parameters
        ----------
        fitness : array_like
            An array of each individual's fitness.
        tournament_size : int
            Number of participants in the tournament.

        Returns
        -------
        int
            The index of the best individual.
        """
        # make next 4 lines prettier and check if PBEA article uses Deterministic tournament selection
        fitness = self.fitness[:,0]
        aspirants = np.random.choice(len(fitness)-1, self.tournament_size, replace=False)
        chosen = []
        for ind in aspirants:
            chosen.append([ind, fitness[ind]])
        chosen.sort(key=lambda x: x[1])

        return chosen[0][0]



import numpy as np
from copy import deepcopy
from random import shuffle


def mate(
    mating_pop, individuals: list, params, crossover_type=None, mutation_type=None
):
    """Swap nodes between two partners and mutate based on standard deviation.

    Parameters
    ----------
    mating_pop : list
        List of indices of individuals to mate. If None, choose from population
        randomly.
        Each entry should contain two indices, one for each parent.
    individuals : list
        List of all individuals.
    params : dict
        Parameters for evolution. If None, use defaults.

    Returns
    -------
    offspring : list
        The offsprings produced as a result of crossover and mutation.
    """

    prob_crossover = params.get("prob_crossover", 0.8)
    prob_mutation = params.get("prob_mutation", 0.3)
    mut_strength = params.get("mut_strength", 1.0)
    cur_gen = params.get("current_total_gen_count", 1)
    total_gen = params.get("total_generations", 10)
    std_dev = (5 / 3) * (1 - cur_gen / total_gen)
    if std_dev < 0:
        std_dev = 0

    if mating_pop is None:
        mating_pop = []
        for i in range(len(individuals)):
            mating_pop.append([i, np.random.randint(len(individuals))])

    offspring = []

    # Gaussian
    if mutation_type == "gaussian" or mutation_type is None:

        for mates in mating_pop:

            offspring1, offspring2 = (
                deepcopy(individuals[mates[0]]),
                deepcopy(individuals[mates[1]]),
            )

            for subnet in range(len(offspring1)):

                sub1 = offspring1[subnet]
                sub2 = offspring2[subnet]

                for layer in range(max(len(sub1), len(sub2))):

                    try:
                        connections = min(sub1[layer].size, sub2[layer].size)

                        # Crossover
                        exchange = np.random.choice(
                            connections,
                            np.random.binomial(connections, prob_crossover),
                            replace=False,
                        )
                        tmp = np.copy(sub1[layer])
                        sub1[layer].ravel()[exchange] = sub2[layer].ravel()[exchange]
                        sub2[layer].ravel()[exchange] = tmp.ravel()[exchange]

                    except IndexError:
                        pass

                    # Mutate the first offspring
                    try:
                        connections = sub1[layer].size

                        mut_val = (
                            np.random.normal(0, std_dev, connections) * mut_strength
                        )

                        mut = np.random.choice(
                            connections,
                            np.random.binomial(connections, prob_mutation),
                            replace=False,
                        )
                        sub1[layer].ravel()[mut] += (
                            sub1[layer].ravel()[mut] * mut_val[mut]
                        )

                    except IndexError:
                        pass

                    # Mutate the second offspring
                    try:
                        connections = sub2[layer].size

                        mut_val = (
                            np.random.normal(0, std_dev, connections) * mut_strength
                        )

                        mut = np.random.choice(
                            connections,
                            np.random.binomial(connections, prob_mutation),
                            replace=False,
                        )
                        sub2[layer].ravel()[mut] += (
                            sub2[layer].ravel()[mut] * mut_val[mut]
                        )

                    except IndexError:
                        continue

            offspring.extend((offspring1, offspring2))

    else:
        pass

    return offspring


class EvoDN2Recombination:
    def __init__(
        self,
        evolver: "BaseEA",
        ProC: float = 0.8,
        ProM: float = 0.3,
        mutation_strength: float = 1.0,
    ):
        self.evolver = evolver
        self.ProC: float = ProC
        self.ProM: float = ProM
        self.mutation_strength: float = mutation_strength

    def do(self, pop, mating_pop_ids: list = None):
        cur_gen = self.evolver.__getattribute__("_current_gen_count")
        total_gen = self.evolver.__getattribute__("total_gen_count")
        pop_size = pop.shape[0]
        if mating_pop_ids is None:
            shuffled_ids = list(range(pop_size))
            shuffle(shuffled_ids)
        else:
            shuffled_ids = mating_pop_ids
        # TODO fix the need for the following
        # [1,2,3,4] -> [[1,2],[3,4]]
        if np.asarray(shuffled_ids).ndim == 1:
            mating_pop = [
                [shuffled_ids[x], shuffled_ids[x]]
                for x in range(len(shuffled_ids))
                if x % 2 == 0
            ]
        elif np.asarray(shuffled_ids).ndim == 2:
            mating_pop = shuffled_ids

        std_dev = (5 / 3) * (1 - cur_gen / total_gen)
        if std_dev < 0:
            std_dev = 0

        offspring = []

        for mates in mating_pop:

            offspring1, offspring2 = (deepcopy(pop[mates[0]]), deepcopy(pop[mates[1]]))

            for subnet in range(len(offspring1)):

                sub1 = offspring1[subnet]
                sub2 = offspring2[subnet]

                for layer in range(max(len(sub1), len(sub2))):

                    try:
                        connections = min(sub1[layer].size, sub2[layer].size)

                        # Crossover
                        exchange = np.random.choice(
                            connections,
                            np.random.binomial(connections, self.ProC),
                            replace=False,
                        )
                        tmp = np.copy(sub1[layer])
                        sub1[layer].ravel()[exchange] = sub2[layer].ravel()[exchange]
                        sub2[layer].ravel()[exchange] = tmp.ravel()[exchange]

                    except IndexError:
                        pass

                    # Mutate the first offspring
                    try:
                        connections = sub1[layer].size

                        mut_val = (
                            np.random.normal(0, std_dev, connections)
                            * self.mutation_strength
                        )

                        mut = np.random.choice(
                            connections,
                            np.random.binomial(connections, self.ProM),
                            replace=False,
                        )
                        sub1[layer].ravel()[mut] += (
                            sub1[layer].ravel()[mut] * mut_val[mut]
                        )

                    except IndexError:
                        pass

                    # Mutate the second offspring
                    try:
                        connections = sub2[layer].size

                        mut_val = (
                            np.random.normal(0, std_dev, connections)
                            * self.mutation_strength
                        )

                        mut = np.random.choice(
                            connections,
                            np.random.binomial(connections, self.ProM),
                            replace=False,
                        )
                        sub2[layer].ravel()[mut] += (
                            sub2[layer].ravel()[mut] * mut_val[mut]
                        )

                    except IndexError:
                        continue

            offspring.extend((offspring1, offspring2))
        return np.asarray(offspring)



from typing import Dict, Type, Tuple, Callable
import numpy as np
import pandas as pd
from desdeo_emo.population.Population import Population
from desdeo_emo.selection.SelectionBase import SelectionBase
from desdeo_problem import MOProblem
from desdeo_emo.EAs import BaseEA
from desdeo_emo.EAs.BaseEA import eaError
from desdeo_tools.interaction import (
   SimplePlotRequest,
   ReferencePointPreference,
    validate_ref_point_with_ideal_and_nadir,
    )


class BaseIndicatorEA(BaseEA):
    """The Base class for indicator based EAs.

    This class contains most of the code to set up the parameters and operators.
    It also contains the logic of a indicator EA.

    Parameters
    ----------
    problem : MOProblem
        The problem class object specifying the details of the problem.
    selection_operator : Type[SelectionBase], optional
        The selection operator to be used by the EA, by default None.
    population_size : int, optional
        The desired population size, by default None, which sets up a default value
        of population size depending upon the dimensionaly of the problem.
    population_params : Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population : Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    a_priori : bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact : bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    n_iterations : int, optional
        The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter : int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations :int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    use_surrogates: bool, optional
    	A bool variable defining whether surrogate problems are to be used or
        not. By default False    
    """

    def __init__(
        self,
        problem: MOProblem,
        population_size: int, # size required
        selection_operator: Type[SelectionBase] = None,
        population_params: Dict = None,
        initial_population: Population = None,
        a_priori: bool = False,
        interact: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        use_surrogates: bool = False,
        # local_fitness np:array for both IBEA and PBEA
        fitnesses = None,
    ):
        super().__init__(
            a_priori=a_priori,
            interact=interact,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
            selection_operator=selection_operator,
            use_surrogates=use_surrogates,
        )

        if initial_population is not None:
            self.population = initial_population
        elif initial_population is None:
            self.population = Population(
                problem, population_size, population_params, use_surrogates
            )
            self._function_evaluation_count += population_size
        

    def end(self):
        """Conducts non-dominated sorting at the end of the evolution process
        Returns:
            tuple: The first element is a 2-D array of the decision vectors of the non-dominated solutions.
                The second element is a 2-D array of the corresponding objective values.
        """
        non_dom = self.population.non_dominated_objectives()
        return (
            self.population.individuals[non_dom, :],
            self.population.objectives[non_dom, :],
        )


    def _next_gen(self):
        """
            Run one generation of indicator based EA. Intended to be used by next_iteration.
        """
        # calls fitness assignment
        self.fitnesses = np.zeros_like(self.population.fitness)
        self.fitnesses = self._fitness_assignment(self.fitnesses)
        # performs the enviromental selection
        while (self.population.pop_size < self.population.individuals.shape[0]):
            worst_index = np.argmin(self.fitnesses, axis=0)[0] # gets the index worst member of population
            self.fitnesses = self._environmental_selection(self.fitnesses, worst_index)
            # remove the worst member from population and from fitnesses
            self.population.delete(worst_index)
            self.fitnesses = np.delete(self.fitnesses, worst_index, 0)

        # perform binary tournament selection. 
        chosen = self._select()
        # variation, call the recombination operators
        offspring = self.population.mate(mating_individuals=chosen)
        self.population.add(offspring)
        self._current_gen_count += 1
        self._gen_count_in_curr_iteration += 1
        self._function_evaluation_count += offspring.shape[0]


    def _select(self) -> list:
        """
            Performs the selection, returns indices of selected individuals. 
            
            Returns
            -------
            list
                List of indices of individuals to be selected.
        """
        return self.selection_operator.do(self.population, self.fitnesses)


    def manage_preferences(self, preference=None):
        """Run the interruption phase of EA.

            Conducts the interaction with the user.
        """
        # start only with reference point reference as in article
        if (self.interact is False): return

        if preference is None:
            msg = "Giving preferences is mandatory"
            raise eaError(msg)

        if not isinstance(preference, ReferencePointPreference):
            msg = (
                f"Wrong object sent as preference. Expected type = "
                f"{type(ReferencePointPreference)}\n"
                f"Recieved type = {type(preference)}"
            )
            raise eaError(msg)

        if preference is not None:
            if preference.request_id != self._interaction_request_id:
                msg = (
                    f"Wrong preference object sent. Expected id = "
                    f"{self._interaction_request_id}.\n"
                    f"Recieved id = {preference.request_id}"
                )
                raise eaError(msg)

        if preference is not None:
            self.reference_point = preference.response.values * self.population.problem._max_multiplier # TODO: there was thing about doing this but i think i do already.
            self.n_iterations += self.n_iterations
            self.total_function_evaluations += self.total_function_evaluations


    def request_preferences(self) -> ReferencePointPreference:
        if (self.interact is False): return None

        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=self.population.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = self.population.problem._max_multiplier
        dimensions_data.loc["ideal"] = self.population.ideal_objective_vector
        dimensions_data.loc["nadir"] = self.population.nadir_objective_vector
        message = ("Please provide preferences as a reference point. Reference point cannot be better than ideal point:\n\n"
            f"{dimensions_data.loc['ideal']}\n"
            f"The reference point will be used to focus the search towards "
            f"the preferred region.\n"
            )

        def validator(dimensions_data: pd.DataFrame, reference_point: pd.DataFrame):
            validate_ref_point_dimensions(dimensions_data, reference_point)
            validate_ref_point_data_type(reference_point)
            validate_ref_point_with_ideal(dimensions_data, reference_point)
            return
                   
        interaction_priority = "required"
        self._interaction_request_id = np.random.randint(0, 1e9)

        return ReferencePointPreference(
                dimensions_data=dimensions_data,
                message=message,
                interaction_priority=interaction_priority,
                preference_validator=validate_ref_point_with_ideal_and_nadir,
                request_id=self._interaction_request_id,
            
        )


    def request_plot(self) -> SimplePlotRequest:
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=self.population.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = self.population.problem._max_multiplier
        dimensions_data.loc["ideal"] = self.population.ideal_objective_vector
        dimensions_data.loc["nadir"] = self.population.nadir_objective_vector
        data = pd.DataFrame(
            self.population.objectives, columns=self.population.problem.objective_names
        )
        return SimplePlotRequest(
            data=data, dimensions_data=dimensions_data, message="Objective Values"
        )


    def requests(self) -> Tuple:
        return (self.request_preferences(), self.request_plot())



from typing import Dict, Callable
import numpy as np
from desdeo_emo.population.Population import Population
from desdeo_emo.selection.TournamentSelection import TournamentSelection
from desdeo_tools.scalarization import SimpleASF
from desdeo_emo.EAs.BaseIndicatorEA import BaseIndicatorEA
from desdeo_tools.utilities.quality_indicator import preference_indicator 
from desdeo_problem import MOProblem


class PBEA(BaseIndicatorEA):
    """Python Implementation of PBEA. 

    Most of the relevant code is contained in the super class. This class just assigns
    the EnviromentalSelection operator to BaseIndicatorEA.

    Parameters
    ----------
    problem: MOProblem
        The problem class object specifying the details of the problem.
    population_size : int, optional
        The desired population size, by default None, which sets up a default value
        of population size depending upon the dimensionaly of the problem.
    population_params : Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population : Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    a_priori : bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact : bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    n_iterations : int, optional
        The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter : int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations : int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    use_surrogates: bool, optional
    	A bool variable defining whether surrogate problems are to be used or
        not. By default False
    kappa : float, optional
        Fitness scaling value for indicators. By default 0.05.
    indicator : Callable, optional
        Quality indicator to use in indicatorEAs. For PBEA this is preference based quality indicator.
    reference_point : np.ndarray
        The reference point that guides the PBEAs search.
    delta : float, optional
        Spesifity for the preference based quality indicator. By default 0.01.

    """
    def __init__(self,
        problem: MOProblem,
        population_size: int,
        initial_population: Population,
        a_priori: bool = False,
        interact: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        use_surrogates: bool = False,
        kappa: float = 0.05,
        indicator: Callable = preference_indicator,
        population_params: Dict = None,
        reference_point = None,
        delta: float = 0.1, 
        ):
        super().__init__(
            problem=problem,
            population_size=population_size,
            population_params=population_params,
            a_priori=a_priori,
            interact=interact,
            n_iterations=n_iterations,
            initial_population=initial_population,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
            use_surrogates=use_surrogates,
        )

        self.kappa = kappa
        self.delta = delta
        self.indicator = indicator
        self.reference_point = reference_point
        selection_operator = TournamentSelection(self.population, 2)
        self.selection_operator = selection_operator


    def _fitness_assignment(self, fitnesses):
        """
            Performs the fitness assignment of the individuals.
        """
        asf = SimpleASF(np.ones_like(self.population.objectives))
        asf_values = asf(self.population.objectives, reference_point=self.reference_point)
        self.min_asf_value = np.min(asf_values)

        for i in range(self.population.individuals.shape[0]):
            for j in range(self.population.individuals.shape[0]):
                if j != i:
                    fitnesses[i] += -np.exp(-self.indicator(self.population.fitness[i], 
                        self.population.fitness[j], self.min_asf_value, self.reference_point, self.delta) / self.kappa)
        return fitnesses


    def _environmental_selection(self, fitnesses, worst_index):
        """
            Selects the worst member of population, then updates the population members fitness values compared to the worst individual.
            Worst individual is removed from the population.
            
        """
        for i in range(self.population.individuals.shape[0]):
            if worst_index != i:
                fitnesses[i] += np.exp(-self.indicator(self.population.fitness[i], 
                    self.population.fitness[worst_index], self.min_asf_value, self.reference_point, 
                    self.delta) / self.kappa)
        return fitnesses




from numpy.core.numeric import indices
from numpy.lib.arraysetops import unique
from pandas.core.frame import DataFrame
from desdeo_tools.scalarization.ASF import SimpleASF, ReferencePointASF
from numba import njit
import numpy as np
import pandas as pd
import copy
from desdeo_problem.surrogatemodels.SurrogateModels import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct,\
    WhiteKernel, RBF, Matern, ConstantKernel
from desdeo_problem import MOProblem



#TODO: add @njit function here
def remove_duplicate(
    X: np.ndarray,
    archive_x: np.ndarray
    ):
    """identifiesthe duplicate rows for decision variables
    Args:
    X (np.ndarray): the current decision variables.
    archive_x (np.ndarray): The decision variables in the archive.

    Returns: 
    indicies (np.ndarray): the indicies of solutions that are NOT already in the archive.
    """

    all_variables = np.vstack((archive_x,X))
    all_variables_indicies = np.arange(len(all_variables))

    e,unique_variables_indicies = np.unique(all_variables, return_index=True, axis=0)

    repeated_all_variables = np.delete(all_variables_indicies, unique_variables_indicies)
    repeated_in_X = repeated_all_variables - len(archive_x)
    X_indicies = np.arange(len(X))
    X_uniqe_indicies = np.delete(X_indicies, repeated_in_X)

        
    return X_uniqe_indicies



def ikrvea_mm(
    reference_point: np.ndarray,
    individuals: np.ndarray,
    objectives: np.ndarray,
    uncertainity: np.ndarray,
    problem: MOProblem,
    u: int) -> float:
    """ Selects the solutions that need to be reevaluated with the original functions.
    This model management is based on the following papaer: 

    'P. Aghaei Pour, T. Rodemann, J. Hakanen, and K. Miettinen, “Surrogate assisted interactive
     multiobjective optimization in energy system design of buildings,” 
     Optimization and Engineering, 2021.'

    Args:
        reference_front (np.ndarray): The reference front that the current front is being compared to.
        Should be an one-dimensional array.
        individuals (np.ndarray): Current individuals generated by using surrogate models
        objectives (np.ndarray): Current objectives  generated by using surrogate models
        uncertainity (np.ndarray): Current Uncertainty values generated by using surrogate models
        problem : the problem class

    Returns:
        float: the new problem object that has an updated archive.
    """     
    
    nd = remove_duplicate(individuals, problem.archive.drop(
            problem.objective_names, axis=1).to_numpy()) #removing duplicate solutions
    if len(nd) == 0:
        return problem
    else:
        non_duplicate_dv = individuals[nd]
        non_duplicate_obj = objectives[nd]
        non_duplicate_unc = uncertainity[nd]
        
    # Selecting solutions with lowest ASF values
    asf_solutions = SimpleASF([1]*problem.n_of_objectives).__call__(non_duplicate_obj, reference_point)
    idx = np.argpartition(asf_solutions, 2*u)
    asf_unc = np.max(non_duplicate_unc [idx[0:2*u]], axis= 1)
    # index of solutions with lowest Uncertainty
    lowest_unc_index = np.argpartition(asf_unc, u)[0:u]
    # evaluating the solutions in asf_unc with lowest uncertainty. The archive will get update in problem.evaluate()
    problem.evaluate(non_duplicate_dv[lowest_unc_index], use_surrogate=False)[0] 
    
    problem.train(models=GaussianProcessRegressor,\
         model_parameters={'kernel': Matern(nu=1.5)}) 

    return problem





import numpy as np

# from pygmo import fast_non_dominated_sorting as nds
from desdeo_tools.utilities import fast_non_dominated_sort
from typing import List
from desdeo_emo.selection.SelectionBase import InteractiveDecompositionSelectionBase
from desdeo_emo.population.Population import Population


class NSGAIII_select(InteractiveDecompositionSelectionBase):
    """The NSGA-III selection operator. Code is heavily based on the version of nsga3 in
        the pymoo package by msu-coinlab.

    Parameters
    ----------
    pop : Population
        [description]
    n_survive : int, optional
        [description], by default None

    """

    def __init__(
        self, pop: Population, n_survive: int = None, selection_type: str = None
    ):
        super().__init__(pop.pop_size, pop.problem.n_of_fitnesses, selection_type)
        self.worst_fitness: np.ndarray = -np.full((1, pop.fitness.shape[1]), np.inf)
        self.extreme_points: np.ndarray = None
        if n_survive is None:
            self.n_survive: int = pop.pop_size
        if selection_type is None:
            selection_type = "mean"
        self.selection_type = selection_type
        self.ideal: np.ndarray = pop.ideal_fitness_val

    def do(self, pop: Population) -> List[int]:
        """Select individuals for mating for NSGA-III.

        Parameters
        ----------
        pop : Population
            The current population.

        Returns
        -------
        List[int]
            List of indices of the selected individuals
        """
        ref_dirs = self.vectors.values_planar
        fitness = self._calculate_fitness(pop)
        # Calculating fronts and ranks
        # fronts, dl, dc, rank = nds(fitness)
        fronts = fast_non_dominated_sort(fitness)
        fronts = [np.where(fronts[i])[0] for i in range(len(fronts))]
        non_dominated = fronts[0]
        fmin = np.amin(fitness, axis=0)
        self.ideal = np.amin(np.vstack((self.ideal, fmin)), axis=0)

        # Calculating worst points
        self.worst_fitness = np.amax(np.vstack((self.worst_fitness, fitness)), axis=0)
        worst_of_population = np.amax(fitness, axis=0)
        worst_of_front = np.max(fitness[non_dominated, :], axis=0)
        self.extreme_points = self.get_extreme_points_c(
            fitness[non_dominated, :], self.ideal, extreme_points=self.extreme_points
        )
        nadir_point = self.get_nadir_point(
            self.extreme_points,
            self.ideal,
            self.worst_fitness,
            worst_of_population,
            worst_of_front,
        )

        # Finding individuals in first 'n' fronts
        selection = np.asarray([], dtype=int)
        for front_id in range(len(fronts)):
            if len(np.concatenate(fronts[: front_id + 1])) < self.n_survive:
                continue
            else:
                fronts = fronts[: front_id + 1]
                selection = np.concatenate(fronts)
                break
        F = fitness[selection]

        last_front = fronts[-1]

        # Selecting individuals from the last acceptable front.
        if len(selection) > self.n_survive:
            niche_of_individuals, dist_to_niche = self.associate_to_niches(
                F, ref_dirs, self.ideal, nadir_point
            )
            # if there is only one front
            if len(fronts) == 1:
                n_remaining = self.n_survive
                until_last_front = np.array([], dtype=np.int)
                niche_count = np.zeros(len(ref_dirs), dtype=np.int)

            # if some individuals already survived
            else:
                until_last_front = np.concatenate(fronts[:-1])
                id_until_last_front = list(range(len(until_last_front)))
                niche_count = self.calc_niche_count(
                    len(ref_dirs), niche_of_individuals[id_until_last_front]
                )
                n_remaining = self.n_survive - len(until_last_front)

            last_front_selection_id = list(range(len(until_last_front), len(selection)))
            if np.any(selection[last_front_selection_id] != last_front):
                print("error!!!")
            selected_from_last_front = self.niching(
                fitness[last_front, :],
                n_remaining,
                niche_count,
                niche_of_individuals[last_front_selection_id],
                dist_to_niche[last_front_selection_id],
            )
            final_selection = np.concatenate(
                (until_last_front, last_front[selected_from_last_front])
            )
            if self.extreme_points is None:
                print("Error")
            if final_selection is None:
                print("Error")
        else:
            final_selection = selection
        return final_selection.astype(int)

    def get_extreme_points_c(self, F, ideal_point, extreme_points=None):
        """Taken from pymoo"""
        # calculate the asf which is used for the extreme point decomposition
        asf = np.eye(F.shape[1])
        asf[asf == 0] = 1e6

        # add the old extreme points to never loose them for normalization
        _F = F
        if extreme_points is not None:
            _F = np.concatenate([extreme_points, _F], axis=0)

        # use __F because we substitute small values to be 0
        __F = _F - ideal_point
        __F[__F < 1e-3] = 0

        # update the extreme points for the normalization having the highest asf value
        # each
        F_asf = np.max(__F * asf[:, None, :], axis=2)
        I = np.argmin(F_asf, axis=1)
        extreme_points = _F[I, :]
        return extreme_points

    def get_nadir_point(
        self,
        extreme_points,
        ideal_point,
        worst_point,
        worst_of_front,
        worst_of_population,
    ):
        LinAlgError = np.linalg.LinAlgError
        try:

            # find the intercepts using gaussian elimination
            M = extreme_points - ideal_point
            b = np.ones(extreme_points.shape[1])
            plane = np.linalg.solve(M, b)
            intercepts = 1 / plane

            nadir_point = ideal_point + intercepts

            if (
                not np.allclose(np.dot(M, plane), b)
                or np.any(intercepts <= 1e-6)
                or np.any(nadir_point > worst_point)
            ):
                raise LinAlgError()

        except LinAlgError:
            nadir_point = worst_of_front

        b = nadir_point - ideal_point <= 1e-6
        nadir_point[b] = worst_of_population[b]
        return nadir_point

    def niching(self, F, n_remaining, niche_count, niche_of_individuals, dist_to_niche):
        survivors = []

        # boolean array of elements that are considered for each iteration
        mask = np.full(F.shape[0], True)

        while len(survivors) < n_remaining:

            # all niches where new individuals can be assigned to
            next_niches_list = np.unique(niche_of_individuals[mask])

            # pick a niche with minimum assigned individuals - break tie if necessary
            next_niche_count = niche_count[next_niches_list]
            next_niche = np.where(next_niche_count == next_niche_count.min())[0]
            next_niche = next_niches_list[next_niche]
            next_niche = next_niche[np.random.randint(0, len(next_niche))]

            # indices of individuals that are considered and assign to next_niche
            next_ind = np.where(
                np.logical_and(niche_of_individuals == next_niche, mask)
            )[0]

            # shuffle to break random tie (equal perp. dist) or select randomly
            np.random.shuffle(next_ind)

            if niche_count[next_niche] == 0:
                next_ind = next_ind[np.argmin(dist_to_niche[next_ind])]
            else:
                # already randomized through shuffling
                next_ind = next_ind[0]

            mask[next_ind] = False
            survivors.append(int(next_ind))

            niche_count[next_niche] += 1

        return survivors

    def associate_to_niches(
        self, F, ref_dirs, ideal_point, nadir_point, utopian_epsilon=0.0
    ):
        utopian_point = ideal_point - utopian_epsilon

        denom = nadir_point - utopian_point
        denom[denom == 0] = 1e-12

        # normalize by ideal point and intercepts
        N = (F - utopian_point) / denom
        dist_matrix = self.calc_perpendicular_distance(N, ref_dirs)

        niche_of_individuals = np.argmin(dist_matrix, axis=1)
        dist_to_niche = dist_matrix[np.arange(F.shape[0]), niche_of_individuals]

        return niche_of_individuals, dist_to_niche

    def calc_niche_count(self, n_niches, niche_of_individuals):
        niche_count = np.zeros(n_niches, dtype=np.int)
        index, count = np.unique(niche_of_individuals, return_counts=True)
        niche_count[index] = count
        return niche_count

    def calc_perpendicular_distance(self, N, ref_dirs):
        u = np.tile(ref_dirs, (len(N), 1))
        v = np.repeat(N, len(ref_dirs), axis=0)

        norm_u = np.linalg.norm(u, axis=1)

        scalar_proj = np.sum(v * u, axis=1) / norm_u
        proj = scalar_proj[:, None] * u / norm_u[:, None]
        val = np.linalg.norm(proj - v, axis=1)
        matrix = np.reshape(val, (len(N), len(ref_dirs)))

        return matrix

    def _calculate_fitness(self, pop) -> np.ndarray:
        if self.selection_type == "mean":
            return pop.fitness
        if self.selection_type == "optimistic":
            return pop.fitness - pop.uncertainity
        if self.selection_type == "robust":
            return pop.fitness + pop.uncertainity



import numpy as np
from typing import List
from desdeo_emo.selection.SelectionBase import InteractiveDecompositionSelectionBase
from desdeo_emo.population.Population import Population
from desdeo_emo.utilities.ReferenceVectors import ReferenceVectors
from desdeo_tools.scalarization.MOEADSF import (
    MOEADSFBase,
)

from scipy.spatial import distance_matrix


class MOEAD_select(InteractiveDecompositionSelectionBase):
    """The MOEAD selection operator.

    Parameters
    ----------
    pop : Population
        The population of individuals
    SF_type : MOEADSFBase
        The scalarizing function employed to evaluate the solutions

    """

    def __init__(
        self,
        pop: Population,
        SF_type: MOEADSFBase,
        n_neighbors: int,
        selection_type: str = None,
    ):
        # initialize
        super().__init__(pop.pop_size, pop.problem.n_of_fitnesses, selection_type)

        self.n_neighbors = n_neighbors
        # Compute the distance between each pair of reference vectors
        distance_matrix_vectors = distance_matrix(
            self.vectors.values_planar, self.vectors.values_planar
        )
        # Get the closest vectors to obtain the neighborhoods
        self.neighborhoods = np.argsort(
            distance_matrix_vectors, axis=1, kind="quicksort"
        )[:, :n_neighbors]
        self.SF_type = SF_type

    def do(
        self,
        pop: Population,
        current_neighborhood: int,
    ) -> List[int]:
        """Select the individuals that are kept in the neighborhood.

        Parameters
        ----------
        pop : Population
            The current population.
        vectors : ReferenceVectors
            Class instance containing reference vectors.
        ideal_point
            Ideal vector found so far
        current_neighborhood
            Neighborhood to be updated
        offspring_fx
            Offspring solution to be compared with the rest of the neighborhood

        Returns
        -------
        List[int]
            List of indices of the selected individuals
        """
        # Compute the value of the SF for each neighbor
        # num_neighbors = len(current_neighborhood) # TODO: check if self.n_neighbors can be used instead.
        current_population = pop.fitness[self.neighborhoods[current_neighborhood], :]
        offspring_fitness = pop.fitness[-1]
        current_reference_vectors = self.vectors.values_planar[
            self.neighborhoods[current_neighborhood], :
        ]
        offspring_population = np.tile(offspring_fitness, (self.n_neighbors, 1))
        ideal_point_matrix = np.tile(pop.ideal_fitness_val, (self.n_neighbors, 1))

        values_SF = self._evaluate_SF(
            current_population, current_reference_vectors, ideal_point_matrix
        )
        values_SF_offspring = self._evaluate_SF(
            offspring_population, current_reference_vectors, ideal_point_matrix
        )

        # Compare the offspring with the individuals in the neighborhood
        # and replace the ones which are outperformed by it.
        selection = np.where(values_SF_offspring < values_SF)[0]

        return self.neighborhoods[current_neighborhood][selection]

    def _evaluate_SF(self, neighborhood, weights, ideal_point):
        # Replace the zeros in the weight vectors with a small value to avoid
        # errors when computing the scalarization function
        fixed_reference_vectors = weights
        fixed_reference_vectors[
            fixed_reference_vectors == 0
        ] = 0.0001  # TODO replace with np.eps?

        # Get the value of the selected scalarization function
        SF_values = np.array(
            list(map(self.SF_type, neighborhood, fixed_reference_vectors, ideal_point))
        )
        return SF_values

    def choose_parents(self, current_neighborhood: int, n_parents: int) -> List[int]:
        current_neighborhood_members = self.neighborhoods[current_neighborhood, :]
        selected_parents = np.random.choice(
            current_neighborhood_members, n_parents, replace=False
        )
        return selected_parents

    def adapt_RVs(self, fitness: np.ndarray) -> None:
        super().adapt_RVs(fitness)
        # Recompute the distance between each pair of reference vectors
        distance_matrix_vectors = distance_matrix(
            self.vectors.values_planar, self.vectors.values_planar
        )
        # Get the closest vectors to obtain the neighborhoods
        self.neighborhoods = np.argsort(
            distance_matrix_vectors, axis=1, kind="quicksort"
        )[:, : self.n_neighbors]



import numpy as np
from random import shuffle


def mate(
    mating_pop, individuals: list, params, crossover_type=None, mutation_type=None
):
    """Swap nodes between two partners and mutate based on standard deviation.

    Parameters
    ----------
    mating_pop : list
        List of indices of individuals to mate. If None, choose from population
        randomly.
        Each entry should contain two indices, one for each parent.
    individuals : list
        List of all individuals.
    params : dict
        Parameters for evolution. If None, use defaults.

    Returns
    -------
    offspring : list
        The offsprings produced as a result of crossover and mutation.
    """

    prob_crossover = params.get("prob_crossover", 0.8)
    prob_mutation = params.get("prob_mutation", 0.3)
    mut_strength = params.get("mut_strength", 1.0)
    cur_gen = params.get("current_total_gen_count", 1)
    total_gen = params.get("total_generations", 10)
    std_dev = (5 / 3) * (1 - cur_gen / total_gen)
    if std_dev < 0:
        std_dev = 0

    if mating_pop is None:
        mating_pop = []
        for i in range(len(individuals)):
            mating_pop.append([i, np.random.randint(len(individuals))])

    offspring = []

    for mates in mating_pop:

        offspring1 = np.copy(individuals[mates[0]])
        offspring2 = np.copy(individuals[mates[1]])

        # Crossover
        for i in range(offspring1.shape[1]):
            if np.random.random() < prob_crossover:
                tmp = np.copy(offspring1[:, i])
                offspring1[:, i] = offspring2[:, i]
                offspring2[:, i] = tmp

        if mutation_type == "gaussian" or mutation_type is None:
            # Method : Gaussian (default)
            # Take a random number of connections based on probability and mutate based
            # on standard deviation, calculated once per generation.

            connections = offspring1.size

            mut_val = np.random.normal(0, std_dev, connections) * mut_strength

            mut = np.random.choice(
                connections,
                np.random.binomial(connections, prob_mutation),
                replace=False,
            )
            offspring1.ravel()[mut] += offspring1.ravel()[mut] * mut_val[mut]

            mut_val = np.random.normal(0, std_dev, connections) * mut_strength

            mut = np.random.choice(
                connections,
                np.random.binomial(connections, prob_mutation),
                replace=False,
            )
            offspring2.ravel()[mut] += offspring2.ravel()[mut] * mut_val[mut]

        elif mutation_type == "self-adapting":
            # Method: Self adapting mutation
            # Choose two random individuals and a random number of connections,
            # mutate offspring based on current gen and connections of two randomly
            # chosen individuals

            # Randomly select two individuals with current match active (=non-zero)
            connections = offspring1.size
            select = np.asarray(individuals)[
                np.random.choice(np.nonzero(np.asarray(individuals))[0], 2)
            ]

            mut = np.random.choice(
                connections,
                np.random.binomial(connections, prob_mutation),
                replace=False,
            )
            offspring1.ravel()[mut] = offspring1.ravel()[mut] + mut_strength * (
                1 - cur_gen / total_gen
            ) * (select[1].ravel()[mut] - select[0].ravel()[mut])

            select = np.asarray(individuals)[
                np.random.choice(np.nonzero(np.asarray(individuals))[0], 2)
            ]

            mut = np.random.choice(
                connections,
                np.random.binomial(connections, prob_mutation),
                replace=False,
            )
            offspring2.ravel()[mut] = offspring2.ravel()[mut] + mut_strength * (
                1 - cur_gen / total_gen
            ) * (select[1].ravel()[mut] - select[0].ravel()[mut])

        else:
            pass

        offspring.extend((offspring1, offspring2))

    return offspring


class EvoNNRecombination:
    def __init__(
        self,
        evolver: "BaseEA",
        ProC: float = 0.8,
        ProM: float = 0.3,
        mutation_strength: float = 1.0,
        mutation_type: str = "gaussian",
    ):
        self.evolver = evolver
        self.mutation_type: str = mutation_type
        self.ProC: float = ProC
        self.ProM: float = ProM
        self.mutation_strength: float = mutation_strength

    def do(self, pop, mating_pop_ids: list = None):
        cur_gen = self.evolver.__getattribute__("_current_gen_count")
        total_gen = self.evolver.__getattribute__("total_gen_count")
        pop_size = pop.shape[0]
        if mating_pop_ids is None:
            shuffled_ids = list(range(pop_size))
            shuffle(shuffled_ids)
        else:
            shuffled_ids = mating_pop_ids
        # TODO fix the need for the following
        # [1,2,3,4] -> [[1,2],[3,4]]
        if np.asarray(shuffled_ids).ndim == 1:
            mating_pop = [
                [shuffled_ids[x], shuffled_ids[x]]
                for x in range(len(shuffled_ids))
                if x % 2 == 0
            ]
        elif np.asarray(shuffled_ids).ndim == 2:
            mating_pop = shuffled_ids
        std_dev = (5 / 3) * (1 - cur_gen / total_gen)
        offspring = []
        for mates in mating_pop:

            offspring1 = np.copy(pop[mates[0]])
            offspring2 = np.copy(pop[mates[1]])

            # Crossover
            for i in range(offspring1.shape[1]):
                if np.random.random() < self.ProC:
                    tmp = np.copy(offspring1[:, i])
                    offspring1[:, i] = offspring2[:, i]
                    offspring2[:, i] = tmp

            if self.mutation_type == "gaussian" or self.mutation_type is None:
                # Method : Gaussian (default)
                # Take a random number of connections based on probability and mutate based
                # on standard deviation, calculated once per generation.

                connections = offspring1.size

                mut_val = (
                    np.random.normal(0, std_dev, connections) * self.mutation_strength
                )

                mut = np.random.choice(
                    connections,
                    np.random.binomial(connections, self.ProM),
                    replace=False,
                )
                offspring1.ravel()[mut] += offspring1.ravel()[mut] * mut_val[mut]

                mut_val = (
                    np.random.normal(0, std_dev, connections) * self.mutation_strength
                )

                mut = np.random.choice(
                    connections,
                    np.random.binomial(connections, self.ProM),
                    replace=False,
                )
                offspring2.ravel()[mut] += offspring2.ravel()[mut] * mut_val[mut]

            elif self.mutation_type == "self-adapting":
                # Method: Self adapting mutation
                # Choose two random individuals and a random number of connections,
                # mutate offspring based on current gen and connections of two randomly
                # chosen individuals

                # Randomly select two individuals with current match active (=non-zero)
                connections = offspring1.size
                select = np.asarray(pop)[
                    np.random.choice(np.nonzero(np.asarray(pop))[0], 2)
                ]

                mut = np.random.choice(
                    connections,
                    np.random.binomial(connections, self.ProM),
                    replace=False,
                )
                offspring1.ravel()[mut] = offspring1.ravel()[
                    mut
                ] + self.mutation_strength * (1 - cur_gen / total_gen) * (
                    select[1].ravel()[mut] - select[0].ravel()[mut]
                )

                select = np.asarray(pop)[
                    np.random.choice(np.nonzero(np.asarray(pop))[0], 2)
                ]

                mut = np.random.choice(
                    connections,
                    np.random.binomial(connections, self.ProM),
                    replace=False,
                )
                offspring2.ravel()[mut] = offspring2.ravel()[
                    mut
                ] + self.mutation_strength * (1 - cur_gen / total_gen) * (
                    select[1].ravel()[mut] - select[0].ravel()[mut]
                )

            else:
                pass

            offspring.extend((offspring1, offspring2))

        return np.asarray(offspring)



from desdeo_problem.problem import ProblemBase, EvaluationResults


class surrogateProblem(ProblemBase):
    def __init__(self, performance_evaluator):
        super().__init__()
        self.performance_evaluator = performance_evaluator
        self.n_of_constraints = 0
        self._max_multiplier = 1
        self.objective_names = ["error", "complexity"]

    def evaluate(self, model_parameters, use_surrogates=False):
        results = self.performance_evaluator(model_parameters)
        return EvaluationResults(results, results, None, None)

    def evaluate_constraint_values(self):
        pass

    def get_variable_bounds(self):
        pass

    def get_objective_names(self):
        return self.objective_names



import numpy as np
from random import shuffle

class SinglePoint_Xover:
    """Simple single point crossover and mutation.

    Parameters
        ----------
        ProC : float, optional
            [description], by default 1
        DisC : float, optional
            [description], by default 30
    """

    def __init__(self, ProC: float = 1, DisC: float = 30):
        """[summary]


        """
        self.ProC = ProC
        self.DisC = DisC

    def crossover(self, ind_0, ind_1):
        """
        Single point crossover.
        Args:
            ind_0: individual_0
            ind_1: individual_1
        Ret:
            new_0, new_1: the individuals generatd after crossover.
        """
        assert(len(ind_0) == len(ind_1))

        point = np.random.randint(len(ind_0))
#         new_0, new_1 = np.zeros(len(ind_0)),  np.zeros(len(ind_0))
        new_0 = np.hstack((ind_0[:point], ind_1[point:]))
        new_1 = np.hstack((ind_1[:point], ind_0[point:]))

        assert(len(new_0) == len(ind_0))
        return new_0, new_1

    def mutation(self, indi):
        """
        Simple mutation.
        Arg:
            indi: individual to mutation.
        """
        point = np.random.randint(len(indi))
        indi[point] = 1 - indi[point]
        return indi

    def do(self, pop: np.ndarray, mating_pop_ids: list = None) -> np.ndarray:
        """Consecutive members of mating_pop_ids are crossed over
            in pairs. Example: if mating_pop_ids = [0, 2, 3, 6, 5] then the individuals
            are crossover as: [0, 2], [3, 6], [5, 0]. Note: if the number of elements
            is odd, the last individual is crossed over with the first one.

        Parameters
        ----------
        pop : np.ndarray
            Array of all individuals
        mating_pop_ids : list, optional
            Indices of population members to mate, by default None, which shuffles and
                mates whole population

        Returns
        -------
        np.ndarray
            The offspring produced as a result of crossover.
        """
        pop_size, num_var = pop.shape
        if mating_pop_ids is None:
            shuffled_ids = list(range(pop_size))
            shuffle(shuffled_ids)
        else:
            shuffled_ids = mating_pop_ids
        mating_pop = pop[shuffled_ids]
        mate_size = len(shuffled_ids)
        if len(shuffled_ids) % 2 == 1:
            # Maybe it should be pop_size-1?
            mating_pop = np.vstack((mating_pop, mating_pop[0]))
            mate_size = mate_size + 1
        # The rest closely follows the matlab code.
        offspring = np.zeros_like(mating_pop)  # empty_like() more efficient?
        for i in range(0, mate_size, 2):

            
            offspring[i], offspring[i + 1] = self.crossover(mating_pop[i],mating_pop[i+1])
            offspring = offspring.round()
        return offspring

class SinglePoint_Mutation:
    """Simple single point crossover and mutation.

    Parameters
        ----------
        ProC : float, optional
            [description], by default 1
        DisC : float, optional
            [description], by default 30
    """

    def __init__(self, ProC: float = 1, DisC: float = 30):
        """[summary]


        """
        self.ProC = ProC
        self.DisC = DisC

    def mutation(self, indi):
        """
        Simple mutation.
        Arg:
            indi: individual to mutation.
        """
        point = np.random.randint(len(indi))
        indi[point] = 1 - indi[point]
        return indi

    def do(self, offspring: np.ndarray):
        """Consecutive members of mating_pop_ids are crossed over
            in pairs. Example: if mating_pop_ids = [0, 2, 3, 6, 5] then the individuals
            are crossover as: [0, 2], [3, 6], [5, 0]. Note: if the number of elements
            is odd, the last individual is crossed over with the first one.

        Parameters
        ----------
        pop : np.ndarray
            Array of all individuals
        mating_pop_ids : list, optional
            Indices of population members to mate, by default None, which shuffles and
                mates whole population

        Returns
        -------
        np.ndarray
            The offspring produced as a result of crossover.
        """
        offspring_size, num_var = offspring.shape
        k = np.random.randint(0,num_var, offspring_size)
       
        for i in range(offspring_size):
            offspring[i,k[i]] = 1 - offspring[i,k[i]]
        
        return offspring



from math import ceil
from typing import Callable, Dict, Type

import numpy as np
import pandas as pd
from desdeo_problem.surrogatemodels.SurrogateModels import BaseRegressor, ModelError
from scipy.special import expit
from sklearn.metrics import mean_squared_error, mean_squared_log_error, r2_score

from desdeo_emo.EAs.BaseEA import BaseEA
from desdeo_emo.EAs.PPGA import PPGA

from desdeo_emo.utilities.plotlyanimate import animate_init_, animate_next_
from desdeo_emo.population.SurrogatePopulation import SurrogatePopulation
from desdeo_emo.recombination.evonn_xover_mutation import EvoNNRecombination
from desdeo_emo.surrogatemodels.Problem import surrogateProblem


def negative_r2_score(y_true, y_pred):
    return -r2_score(y_true, y_pred)


class EvoNN(BaseRegressor):
    def __init__(
        self,
        num_hidden_nodes: int = 20,
        p_omit: float = 0.2,
        w_low: float = -5.0,
        w_high: float = 5.0,
        activation_function: str = "sigmoid",
        loss_function: str = "mse",
        training_algorithm: Type[BaseEA] = PPGA,
        pop_size: int = 500,  # Add ProC and ProM
        model_selection_criterion: str = "akaike_corrected",
        recombination_type: str = "evonn_xover_mutation",
        crossover_type: str = "standard",
        mutation_type: str = "gaussian",
    ):
        loss_functions = {
            "mse": mean_squared_error,
            "msle": mean_squared_log_error,
            "neg_r2": negative_r2_score,
        }
        # Hyperparameters
        self.num_hidden_nodes: int = num_hidden_nodes
        self.p_omit: float = p_omit
        self.w_low: float = w_low
        self.w_high: float = w_high
        self.activation_function: str = activation_function
        self.loss_function_str: str = loss_function
        self.loss_function: Callable = loss_functions[loss_function]
        self.training_algorithm: Type[BaseEA] = training_algorithm
        self.pop_size: int = pop_size
        self.model_selection_criterion: str = model_selection_criterion
        self.recombination_type: str = recombination_type
        self.crossover_type: str = crossover_type
        self.mutation_type: str = mutation_type
        self.X: np.ndarray = None
        self.y: np.ndarray = None
        self.model_trained: bool = False
        # Model Parameters
        self._first_layer: np.ndarray = None
        self._last_layer: np.ndarray = None
        # Extras
        self.performance: Dict = {"RMSE": None, "R^2": None, "AICc": None}
        self.model_population = None

    def fit(self, X: np.ndarray, y: np.ndarray):
        if isinstance(X, (pd.DataFrame, pd.Series)):
            X = X.values
        if isinstance(y, (pd.DataFrame, pd.Series)):
            y = y.values.reshape(-1, 1)
        if X.shape[0] != y.shape[0]:
            msg = (
                f"Ensure that the number of samples in X and y are the same"
                f"Number of samples in X = {X.shape[0]}"
                f"Number of samples in y = {y.shape[0]}"
            )
            raise ModelError(msg)
        self.X = X
        self.y = y

        # Create problem
        problem = surrogateProblem(performance_evaluator=self._model_performance)
        problem.n_of_objectives = 2
        # Create Population
        initial_pop = self._create_individuals()
        population = SurrogatePopulation(
            problem, self.pop_size, initial_pop, None, None, None
        )
        # Do evolution
        evolver = self.training_algorithm(problem, initial_population=population)
        recombinator = EvoNNRecombination(
            evolver=evolver, mutation_type=self.mutation_type
        )
        evolver.population.recombination = recombinator
        figure = animate_init_(evolver.population.objectives, filename="EvoNN.html")
        while evolver.continue_evolution():
            evolver.iterate()
            figure = animate_next_(
                evolver.population.objectives,
                figure,
                filename="EvoNN.html",
                generation=evolver._iteration_counter,
            )
        self.model_population = evolver.population
        # Selection
        self.select()
        self.model_trained = True

    def _model_performance(
        self,
        first_layer: np.ndarray = None,
        X: np.ndarray = None,
        y_true: np.ndarray = None,
    ):
        if first_layer is None and self.model_trained is False:
            msg = "Model has not been trained yet"
            raise ModelError(msg)
        if first_layer is None:
            first_layer = self._first_layer
        if X is None:
            X = self.X
            y = self.y
        if np.ndim(first_layer) == 3:
            loss = []
            complexity = []
            for actual_first_layer in first_layer:
                y_predict = self.predict(
                    X=X, first_layer=actual_first_layer, training=True
                )
                loss.append(self.loss_function(y, y_predict))
                complexity.append(np.count_nonzero(actual_first_layer))
        elif np.ndim(first_layer) == 2:
            y_predict = self.predict(X=X, first_layer=first_layer, training=True)
            loss = self.loss_function(y, y_predict)
            complexity = np.count_nonzero(first_layer)
        return np.asarray((loss, complexity)).T

    def predict(
        self,
        X: np.ndarray = None,
        first_layer: np.ndarray = None,
        training: bool = False,
    ):
        if first_layer is None and self.model_trained is False:
            msg = "Model has not been trained yet"
            raise ModelError(msg)
        elif first_layer is not None:
            # Calculate the dot product + bias
            out = np.dot(X, first_layer[1:, :]) + first_layer[0]
            activated_layer = self.activate(out)
            last_layer = self.calculate_linear(activated_layer)
        elif first_layer is None:
            first_layer = self._first_layer
            last_layer = self._last_layer
            # Calculate the dot product + bias
            out = np.dot(X, first_layer[1:, :]) + first_layer[0]
            activated_layer = self.activate(out)
        else:
            msg = "How did you get here?"
            raise ModelError(msg)
        y_pred = np.dot(activated_layer, last_layer[1:, :]) + last_layer[0]
        if training:
            return y_pred
        else:
            return y_pred[:, 0], np.zeros_like(y_pred[:, 0])

    def activate(self, x):
        if self.activation_function == "sigmoid":
            return expit(x)
        elif self.activation_function == "relu":
            return np.maximum(x, 0)
        elif self.activation_function == "tanh":
            return np.tanh(x)
        else:
            msg = (
                f"Given activation function not recognized: {self.activation_function}"
                f"\nActivation function should be one of ['relu', 'sigmoid', 'tanh']"
            )
            raise ModelError(msg)

    def calculate_linear(self, previous_layer_output):
        """Calculate the final layer using LLSQ or

        Parameters
        ----------
        non_linear_layer : np.ndarray
            Output of the activation function

        Returns
        -------
        linear_layer : np.ndarray
            The optimized weight matrix of the upper part of the network
        predicted_values : np.ndarray
            The prediction of the model
        training_error : float
            The model's training error
        """

        linear_layer = None
        previous_layer_output = np.hstack(
            (np.ones((previous_layer_output.shape[0], 1)), previous_layer_output)
        )
        linear_solution = np.linalg.lstsq(previous_layer_output, self.y, rcond=None)
        linear_layer = linear_solution[0]
        return linear_layer

    def _create_individuals(self):

        individuals = np.random.uniform(
            self.w_low,
            self.w_high,
            size=(self.pop_size, self.X.shape[1], self.num_hidden_nodes),
        )

        # Set bias
        individuals = np.insert(individuals, 0, 1, axis=1)

        # Randomly set some weights to zero
        zeros = np.random.choice(
            np.arange(individuals.size), ceil(individuals.size * self.p_omit)
        )
        individuals.ravel()[zeros] = 0
        return individuals

    def select(self):
        aicc_array = []
        if self.model_selection_criterion == "min_error":
            # Return the model with the lowest error
            selected = np.argmin(self.model_population.objectives[:, 0])
        elif self.model_selection_criterion == "akaike_corrected":
            for first_layer in self.model_population.individuals:
                # Blah
                out = np.dot(self.X, first_layer[1:, :]) + first_layer[0]
                activated_layer = self.activate(out)
                last_layer = self.calculate_linear(activated_layer)
                y_pred = np.dot(activated_layer, last_layer[1:, :]) + last_layer[0]
                # Blah 2
                k = np.count_nonzero(first_layer) + np.count_nonzero(last_layer)
                rss = np.sum((y_pred - self.y) ** 2)
                num_samples = self.y.shape[0]
                aic = 2 * k + num_samples * np.log(rss / num_samples)
                aicc = aic + (2 * k * (k + 1)) / (num_samples - k - 1)
                aicc_array.append(aicc)
            selected = np.argmin(aicc)
        # Blah 3
        first_layer = self.model_population.individuals[selected]
        out = np.dot(self.X, first_layer[1:, :]) + first_layer[0]
        activated_layer = self.activate(out)
        last_layer = self.calculate_linear(activated_layer)
        y_pred = np.dot(activated_layer, last_layer[1:, :]) + last_layer[0]
        # Blah4
        self._first_layer = first_layer
        self._last_layer = last_layer
        self.performance["RMSE"] = np.sqrt(mean_squared_error(self.y, y_pred))
        self.performance["R^2"] = r2_score(self.y, y_pred)
        self.performance["AICc"] = aicc_array[selected]



from abc import ABC, abstractmethod
from typing import List, Union

import numpy as np
import pandas as pd
from desdeo_emo.population.Population import Population
from desdeo_emo.utilities.ReferenceVectors import ReferenceVectors
from desdeo_tools.interaction import (
    BoundPreference,
    NonPreferredSolutionPreference,
    PreferredSolutionPreference,
    ReferencePointPreference,
    validate_ref_point_data_type,
    validate_ref_point_dimensions,
    validate_ref_point_with_ideal,
)


class SelectionBase(ABC):
    """The base class for the selection operator.
    """

    @abstractmethod
    def do(self, pop: Population, *args) -> List[int]:
        """Use the selection operator over the given fitness values. Return the indices
            individuals with the best fitness values according to the operator.

        Parameters
        ----------
        fitness : np.ndarray
            Fitness of the individuals from which the next generation is to be selected.

        Returns
        -------
        List[int]
            The list of selected individuals
        """


class InteractiveDecompositionSelectionBase(SelectionBase):
    """The base class for the selection operators for interactive decomposition based MOEAs.
    """

    def __init__(
        self,
        number_of_vectors: int,
        number_of_objectives: int,
        selection_type: str = None,
    ):
        self.vectors = ReferenceVectors(
            number_of_vectors=number_of_vectors,
            number_of_objectives=number_of_objectives,
        )
        if selection_type is None:
            selection_type = "mean"
        self.selection_type = selection_type
        self._interaction_request_id: int = None
        self.allowable_interaction_types = {
            "Reference point": (
                "Specify a reference point worse than the utopian point. "
                "The reference vectors are focused around the vector joining "
                "provided reference point and the utopian point. "
                "New solutions are searched for in this focused region of interest."
            ),
            "Preferred solutions": (
                "Choose one or more solutions as the preferred solutions. "
                "The reference vectors are focused around the vector joining "
                "the utopian point and the preferred solutions. "
                "New solutions are searched for in this focused regions of interest."
            ),
            "Non-preferred solutions": (
                "Choose one or more solutions that are not preferred. "
                "The reference vectors near such solutions are removed. "
                "New solutions are hence not searched for in areas close to these solutions."
            ),
            "Preferred ranges": (
                "Provide preferred values for the upper and lower bounds of all objectives. "
                "New reference vectors are generated within these bounds. "
                "New solutions are searched for in this bounded region of interest."
            ),
        }
        self.interaction_type: str = None

    def _calculate_fitness(self, pop: Population) -> np.ndarray:
        if self.selection_type == "mean":
            return pop.fitness
        if self.selection_type == "optimistic":
            return pop.fitness - pop.uncertainity
        if self.selection_type == "robust":
            return pop.fitness + pop.uncertainity

    def set_interaction_type(self, interaction_type: str = None) -> Union[None, dict]:
        if interaction_type is None:
            return self.allowable_interaction_types
        elif interaction_type not in self.allowable_interaction_types.keys():
            raise ValueError(
                f"Invalid interaction type. Interaction type must be one of {self.allowable_interaction_types.keys()}."
            )
        else:
            self.interaction_type = interaction_type

    def adapt_RVs(self, fitness: np.ndarray) -> None:
        self.vectors.adapt(fitness)
        self.vectors.neighbouring_angles()

    def request_preferences(
        self, pop: Population
    ) -> Union[
        PreferredSolutionPreference,
        NonPreferredSolutionPreference,
        ReferencePointPreference,
        BoundPreference,
    ]:
        if self.interaction_type == "Reference point":
            return self.request_reference_point(pop)
        if self.interaction_type == "Preferred solutions":
            return self.request_preferred_solutions(pop)
        if self.interaction_type == "Non-preferred solutions":
            return self.request_non_preferred_solutions(pop)
        if self.interaction_type == "Preferred ranges":
            return self.request_preferred_ranges(pop)
        raise ValueError("Interaction type not set.")

    def request_reference_point(self, pop: Population) -> ReferencePointPreference:
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=pop.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = pop.problem._max_multiplier
        dimensions_data.loc["ideal"] = pop.ideal_objective_vector
        dimensions_data.loc["nadir"] = pop.nadir_objective_vector
        message = (
            f"Please provide a reference point worse than the ideal point:\n\n"
            f"{dimensions_data.loc['ideal']}\n"
            f"The reference point will be used to focus the reference vectors towards "
            f"the preferred region.\n"
            f"If a reference point is not provided, the reference vectors are spread uniformly"
            f" in the objective space."
        )

        def validator(dimensions_data: pd.DataFrame, reference_point: pd.DataFrame):
            validate_ref_point_dimensions(dimensions_data, reference_point)
            validate_ref_point_data_type(reference_point)
            validate_ref_point_with_ideal(dimensions_data, reference_point)
            return

        self._interaction_request_id = np.random.randint(0, 1e9)

        return ReferencePointPreference(
            dimensions_data=dimensions_data,
            message=message,
            interaction_priority="recommended",
            preference_validator=validator,
            request_id=self._interaction_request_id,
        )

    def request_preferred_solutions(
        self, pop: Population
    ) -> PreferredSolutionPreference:
        message = (
            "Please specify index/indices of preferred solutions in a numpy array (indexing starts from 0).\n"
            "For example: \n"
            "\tnumpy.array([1]), for choosing the solutions with index 1.\n"
            "\tnumpy.array([2, 4, 5, 16]), for choosing the solutions with indices 2, 4, 5, and 16.\n"
            "The reference vectors will be focused around the chosen preferred solutions."
        )

        self._interaction_request_id = np.random.randint(0, 1e9)

        return PreferredSolutionPreference(
            n_solutions=pop.objectives.shape[0],
            message=message,
            interaction_priority="recommended",
            request_id=self._interaction_request_id,
        )

    def request_non_preferred_solutions(
        self, pop: Population
    ) -> NonPreferredSolutionPreference:

        message = (
            "Please specify index/indices of non-preferred solutions in a numpy array (indexing starts from 0).\n"
            "For example: \n"
            "\tnumpy.array([3]), for choosing the solutions with index 3.\n"
            "\tnumpy.array([1, 2]), for choosing the solutions with indices 1 and 2.\n"
            "The reference vectors are focused away from the non*åreferred solutions."
        )

        self._interaction_request_id = np.random.randint(0, 1e9)
        return NonPreferredSolutionPreference(
            n_solutions=pop.objectives.shape[0],
            message=message,
            interaction_priority="recommended",
            request_id=self._interaction_request_id,
        )

    def request_preferred_ranges(self, pop) -> BoundPreference:
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=pop.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = pop.problem._max_multiplier
        dimensions_data.loc["ideal"] = pop.ideal_objective_vector
        dimensions_data.loc["nadir"] = pop.nadir_objective_vector
        message = (
            "Please specify desired lower and upper bound for each objective, starting from \n"
            "the first objective and ending with the last one. Please specify the bounds as a numpy array containing \n"
            "lists, so that the first item of list is the lower bound and the second the upper bound, for each \n"
            "objective. \n"
            "The reference vectors are regenerated within the preferred ranges.\n"
            "\tFor example: numpy.array([[1, 2], [2, 5], [0, 3.5]]), for a problem with three "
            "objectives.\n"
            f"Ideal vector: \n{dimensions_data.loc['ideal']}\n"
            f"Nadir vector: \n{dimensions_data.loc['nadir']}."
        )
        self._interaction_request_id = np.random.randint(0, 1e9)

        BoundPreference(
            dimensions_data=dimensions_data,
            n_objectives=pop.problem.n_of_objectives,
            message=message,
            interaction_priority="recommended",
            request_id=self._interaction_request_id,
        )

    def manage_preferences(
        self,
        pop: Population,
        preference: Union[
            PreferredSolutionPreference,
            NonPreferredSolutionPreference,
            ReferencePointPreference,
            BoundPreference,
            None,
        ],
    ):

        if preference is None:
            return self.adapt_RVs(pop.fitness)
        if self._interaction_request_id != preference.request_id:
            raise ValueError("Wrong request object provided. Request IDs don't match.")
        if self.interaction_type == "Reference point":
            return self.manage_reference_point(pop, preference)
        if self.interaction_type == "Preferred solutions":
            return self.manage_preferred_solutions(pop, preference)
        if self.interaction_type == "Non-preferred solutions":
            return self.manage_non_preferred_solutions(pop, preference)
        if self.interaction_type == "Preferred ranges":
            return self.manage_preferred_ranges(preference)
        raise ValueError("Interaction type not set.")

    def manage_reference_point(
        self, pop: Population, preference: ReferencePointPreference
    ):
        if not isinstance(preference, ReferencePointPreference):
            raise TypeError(
                "Preference object must be an instance of ReferencePointPreference."
            )
        ideal = pop.ideal_fitness_val
        refpoint = preference.response.values * pop.problem._max_multiplier
        refpoint = refpoint - ideal
        norm = np.sqrt(np.sum(np.square(refpoint)))
        refpoint = refpoint / norm
        self.vectors.iteractive_adapt_3(refpoint)
        self.vectors.add_edge_vectors()
        self.vectors.neighbouring_angles()

    def manage_preferred_solutions(
        self, pop: Population, preference: PreferredSolutionPreference
    ):
        if not isinstance(preference, PreferredSolutionPreference):
            raise TypeError(
                "Preference object must be an instance of PreferredSolutionPreference."
            )
        self.vectors.interactive_adapt_1(
            z=pop.objectives[preference.response],
            n_solutions=np.shape(pop.objectives)[0],
        )
        self.vectors.add_edge_vectors()
        self.vectors.neighbouring_angles()

    def manage_non_preferred_solutions(
        self, pop: Population, preference: NonPreferredSolutionPreference
    ):
        if not isinstance(preference, NonPreferredSolutionPreference):
            raise TypeError(
                "Preference object must be an instance of NonPreferredSolutionPreference."
            )
        self.vectors.interactive_adapt_2(
            z=pop.objectives[preference.response],
            n_solutions=np.shape(pop.objectives)[0],
        )
        self.vectors.add_edge_vectors()
        self.vectors.neighbouring_angles()

    def manage_preferred_ranges(self, preference: BoundPreference):
        if not isinstance(preference, BoundPreference):
            raise TypeError("Preference object must be an instance of BoundPreference.")
        self.vectors.interactive_adapt_4(preference.response)
        self.vectors.add_edge_vectors()
        self.vectors.neighbouring_angles()



"""This module provides implementations of various selection operators.
"""
__all__ = ["APD_Select", "NSGAIII_select", "TournamentSelection", "MOEAD_select"]

from desdeo_emo.selection.APD_Select_constraints import APD_Select
from desdeo_emo.selection.NSGAIII_select import NSGAIII_select
from desdeo_emo.selection.TournamentSelection import TournamentSelection 
from desdeo_emo.selection.MOEAD_select import MOEAD_select



from desdeo_emo.population.Population import Population
from desdeo_emo.EAs.BaseEA import BaseEA, eaError
from desdeo_emo.selection.TournamentSelection import TournamentSelection
import numpy as np


class TournamentEA(BaseEA):
    def __init__(
        self,
        problem,
        initial_population: Population,
        n_gen_per_iter: int = 10,
        n_iterations: int = 10,
        tournament_size: int = 5,
        population_size: int = 500,
    ):
        """Run generations of evolutionary algorithm using tournament selection.

        Parameters
        ----------
        population : "Population"
            This variable is updated as evolution takes place
        ea_parameters : dict
            Takes the EA parameters.

        Returns
        -------
        Population:
            Returns the Population after evolution.
        """
        super().__init__(n_gen_per_iter=n_gen_per_iter, n_iterations=n_iterations)
        if initial_population is None:
            msg = "Provide initial population"
            raise eaError(msg)
        self.population = initial_population
        self.target_pop_size = population_size
        self.tournament_size = tournament_size
        selection_operator = TournamentSelection(self.population, self.tournament_size)
        self.selection_operator = selection_operator


    def start(self):
        pass

    def end(self):
        """Conducts non-dominated sorting at the end of the evolution process
        Returns:
            tuple: The first element is a 2-D array of the decision vectors of the non-dominated solutions.
                The second element is a 2-D array of the corresponding objective values.
        """
        non_dom = self.population.non_dominated_objectives()
        return (
            self.population.individuals[non_dom, :],
            self.population.objectives[non_dom, :],
        )

    def _next_gen(self):
        """Run one generation of decomposition based EA.

        This method leaves method.params unchanged. Intended to be used by
        next_iteration.

        Parameters
        ----------
        population: "Population"
            Population object
        """

        selected = self.select()
        offspring = self.population.mate(mating_individuals=selected)
        self.population.delete(np.arange(len(self.population.individuals)))
        self.population.add(offspring)
        self._current_gen_count += 1
        self._gen_count_in_curr_iteration += 1
        self._function_evaluation_count += offspring.shape[0]

    def select(self) -> list:
       return self.selection_operator.do(self.population)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['desdeo_emo',
 'desdeo_emo.EAs',
 'desdeo_emo.population',
 'desdeo_emo.problem',
 'desdeo_emo.recombination',
 'desdeo_emo.selection',
 'desdeo_emo.surrogatemodels',
 'desdeo_emo.utilities']

package_data = \
{'': ['*']}

install_requires = \
['desdeo-problem>=1.4,<2.0', 'plotly>=5.10.0,<6.0.0', 'pyDOE>=0.3.8,<0.4.0']

setup_kwargs = {
    'name': 'desdeo-emo',
    'version': '1.4.2',
    'description': 'The python version reference vector guided evolutionary algorithm.',
    'long_description': '# desdeo-emo\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/industrial-optimization-group/desdeo-emo/master)\n\nThe evolutionary algorithms package within the [DESDEO framework](https://github.com/industrial-optimization-group/DESDEO).\n\nCode for the SoftwareX paper can be found in [this notebook](docs/notebooks/Using_EvoNN_for_optimization.ipynb).\n\nCurrently supported:\n* Multi-objective optimization with visualization and interaction support.\n* Preference is accepted as a reference point.\n* Surrogate modelling (neural networks and genetic trees) evolved via EAs.\n* Surrogate assisted optimization\n* Constraint handling using `RVEA`\n* IOPIS optimization using `RVEA` and `NSGA-III`\n\nCurrently _NOT_ supported:\n* Binary and integer variables.\n\nTo test the code, open the [binder link](https://mybinder.org/v2/gh/industrial-optimization-group/desdeo-emo/master) and read example.ipynb.\n\nRead the documentation [here](https://desdeo-emo.readthedocs.io/en/latest/)\n\n### Requirements\n* Python 3.8 or newer.\n* [Poetry dependency manager](https://github.com/sdispater/poetry): Only for developers\n\n### Installation process for normal users\n* Create a new virtual enviroment for the project\n* Run: `pip install desdeo_emo`\n\n### Installation process for developers\n* Download and extract the code or `git clone`\n* Create a new virtual environment for the project\n* Run `poetry install` inside the virtual environment shell.\n\n## Citation\n\nIf you decide to use DESDEO is any of your works or research, we would appreciate you citing the appropiate paper published in [IEEE Access](https://doi.org/10.1109/ACCESS.2021.3123825) (open access).\n',
    'author': 'Bhupinder Saini',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<3.10',
}


setup(**setup_kwargs)



import numpy as np
from random import choice, sample


def mutate(offspring, individuals, params, *args):
    """Perform BioGP mutation functions.

    Standard mutation:
    Randomly select and regrow a subtree of an individual.

    Small mutation:
    Randomly select a node within a tree and replace it with either a function of the
    same arity,
    or another value from the terminal set.

    Mono parental:
    Randomly swap two subtrees within the same individual.

    Parameters
    ----------
    offspring : list
        List of individuals to mutate.
    individuals : list
        List of all individuals.
    params : dict
        Parameters for breeding. If None, use defaults.

    """

    prob_mut = params.get("prob_mutation", 0.3)
    prob_stand = 1 / 3 * prob_mut
    prob_point = 1 / 3 * prob_mut
    prob_mono = prob_mut - prob_stand - prob_point
    prob_replace = prob_mut
    r = np.random.rand()

    for ind in offspring:
        if r <= prob_stand:
            # Standard mutation
            #
            # This picks a random subtree anywhere within the tree
            rand_node = choice(ind.nodes[1:])
            tree = ind.grow_tree(method="grow", depth=rand_node.depth, ind=rand_node)
            rand_node.value = tree.value
            rand_node.roots = tree.roots

            # This picks a whole subtree at depth=1 under the linear node
            # rand_subtree = np.random.randint(len(ind.roots))
            # del ind.roots[rand_subtree]
            # ind.grow_tree(method="grow", ind=ind)

            ind.nodes = ind.get_sub_nodes()

        elif r <= prob_point + prob_stand:
            # Small mutation
            for node in ind.nodes[1:]:
                if np.random.rand() < prob_replace and callable(node.value):
                    value = choice(node.function_set)
                    while node.value.__code__.co_argcount != value.__code__.co_argcount:
                        value = choice(node.function_set)
                    node.value = value
                elif np.random.rand() < prob_replace:
                    node.value = choice(node.terminal_set)
            ind.nodes = ind.get_sub_nodes()

        elif r <= prob_mono + prob_point + prob_stand:
            # Mono parental
            swap_nodes = sample(ind.nodes[1:], 2)
            tmp_value = swap_nodes[0].value
            tmp_roots = swap_nodes[0].roots
            swap_nodes[0].value = swap_nodes[1].value
            swap_nodes[0].roots = swap_nodes[1].roots
            swap_nodes[1].value = tmp_value
            swap_nodes[1].roots = tmp_roots
            ind.nodes = ind.get_sub_nodes()

        else:
            pass


class BioGP_mutation:
    def __init__(self, probability_mutation: float):
        self.probability_mutation: float = probability_mutation
        self.probability_stand = probability_mutation / 3
        self.probability_point = probability_mutation / 3
        self.probability_mono = probability_mutation / 3
        self.probability_replace = probability_mutation

    def do(self, offspring):
        r = np.random.rand()

        for ind in offspring[:, 0]:
            if r <= self.probability_stand:
                # Standard mutation
                #
                # This picks a random subtree anywhere within the tree
                rand_node = choice(ind.nodes[1:])
                tree = ind.grow_tree(
                    method="grow", depth=rand_node.depth, ind=rand_node
                )
                rand_node.value = tree.value
                rand_node.roots = tree.roots

                # This picks a whole subtree at depth=1 under the linear node
                # rand_subtree = np.random.randint(len(ind.roots))
                # del ind.roots[rand_subtree]
                # ind.grow_tree(method="grow", ind=ind)

                ind.nodes = ind.get_sub_nodes()

            elif r <= self.probability_point + self.probability_stand:
                # Small mutation
                for node in ind.nodes[1:]:
                    if np.random.rand() < self.probability_replace and callable(
                        node.value
                    ):
                        value = choice(node.function_set)
                        while (
                            node.value.__code__.co_argcount
                            != value.__code__.co_argcount
                        ):
                            value = choice(node.function_set)
                        node.value = value
                    elif np.random.rand() < self.probability_replace:
                        node.value = choice(node.terminal_set)
                ind.nodes = ind.get_sub_nodes()

            elif r <= self.probability_mutation:
                # Mono parental
                swap_nodes = sample(ind.nodes[1:], 2)
                tmp_value = swap_nodes[0].value
                tmp_roots = swap_nodes[0].roots
                swap_nodes[0].value = swap_nodes[1].value
                swap_nodes[0].roots = swap_nodes[1].roots
                swap_nodes[1].value = tmp_value
                swap_nodes[1].roots = tmp_roots
                ind.nodes = ind.get_sub_nodes()

            else:
                pass

        return np.asarray(offspring)



from typing import Tuple

import numpy as np
from desdeo_tools.utilities import dominates
from numba import njit


@njit()
def check_domination(
    new: np.ndarray, archive: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Checks if newly evaluated and archived solutions are non_dominated.

    Args:
        new (np.ndarray): Newly evaluated solution's fitness values
             (should be a 2D array).
        archive (np.ndarray): Archive of non-dominated solutions
             (also a 2D array, must be mutually non-dominated).

    Raises:
        ValueError: If the new array is not a 2D array.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Two arrays of indices that are
             non-dominated in the new and archive arrays respectively.
    """
    if new.ndim != 2:
        raise ValueError("New array should have 2 dimensions.")
    non_dominated_in_new = np.ones(len(new), dtype=np.bool_)
    non_dominated_in_archive = np.ones(len(archive), dtype=np.bool_)
    for i in range(len(new)):
        for j in range(len(archive)):
            if not non_dominated_in_archive[j]:
                continue
            if dominates(archive[j], new[i]):
                non_dominated_in_new[i] = False
                break  # No need to check with the rest of the archive
            if dominates(new[i], archive[j]):
                non_dominated_in_archive[j] = False
    return (non_dominated_in_new, non_dominated_in_archive)



from typing import Dict

from desdeo_emo.EAs.BaseEA import BaseDecompositionEA
from desdeo_emo.population.Population import Population
from desdeo_emo.selection.MOEAD_select import MOEAD_select
from desdeo_problem import MOProblem
from desdeo_problem.problem.Problem import EvaluationResults
from desdeo_tools.scalarization import MOEADSF
from desdeo_tools.scalarization.MOEADSF import PBI


class MOEA_D(BaseDecompositionEA):
    """Python implementation of MOEA/D

    .. Q. Zhang and H. Li, "MOEA/D: A Multiobjective Evolutionary Algorithm Based on Decomposition,"
    in IEEE Transactions on Evolutionary Computation, vol. 11, no. 6, pp. 712-731, Dec. 2007, doi: 10.1109/TEVC.2007.892759.

    Parameters
    ----------
    problem: MOProblem
        The problem class object specifying the details of the problem.
    scalarization_function: MOEADSF
        The scalarization function to compare the solutions. Some implementations
        can be found in desdeo-tools/scalarization/MOEADSF. By default it uses the
        PBI function.
    n_neighbors: int, optional
        Number of reference vectors considered in the neighborhoods creation. The default
        number is 20.
    population_params: Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population: Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    lattice_resolution: int, optional
        The number of divisions along individual axes in the objective space to be
        used while creating the reference vector lattice by the simplex lattice
        design. By default None
    n_parents: int, optional
        Number of individuals considered for the generation of offspring solutions. The default
        option is 2.
    a_priori: bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact: bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    use_surrogates: bool, optional
        A bool variable defining whether surrogate problems are to be used or
        not. By default False
    n_iterations: int, optional
         The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter: int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations: int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    """

    def __init__(  # parameters of the class
        self,
        problem: MOProblem,
        n_neighbors: int = 20,
        n_parents: int = 2,
        use_repair: bool = True,
        initial_population: Population = None,
        population_size: int = None,
        population_params: Dict = None,
        lattice_resolution: int = None,
        interact: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        use_surrogates: bool = False,
        total_function_evaluations: int = 0,
        scalarization_function: MOEADSF = PBI(),
        keep_archive: bool = False,
        save_non_dominated: bool = False,
    ):
        super().__init__(  # parameters for decomposition based approach
            problem=problem,
            population_params=population_params,
            population_size=population_size,
            keep_archive=keep_archive,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            interact=interact,
            use_surrogates=use_surrogates,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
            save_non_dominated=save_non_dominated,
        )

        self.use_repair = use_repair

        # New crossover and mutation operators needed for more than two parents.
        if n_parents != 2:
            raise ValueError(
                "n_parents must be equal to 2. Other values not yet supported."
            )
        self.n_parents = n_parents

        self.selection_operator = MOEAD_select(
            self.population, SF_type=scalarization_function, n_neighbors=n_neighbors
        )

        # self._ideal_point = self.population.ideal_fitness_val

    def _next_gen(self):
        # For each individual from the population
        for i in range(self.population.pop_size):
            # Consider only the individuals of the current neighborhood
            # for parent selection
            selected_parents = self.selection_operator.choose_parents(
                current_neighborhood=i, n_parents=self.n_parents
            )

            # TODO: Implement better recombination operators for this steps
            # Currently SBX and PolyM are used, they produce two offspring.
            # Apply genetic operators over two random individuals
            offspring = self.population.mate(selected_parents)
            offspring = offspring[0, :]

            # Repair the solution if it is needed
            if self.use_repair:
                offspring = self.population.repair(offspring)

            # Evaluate the offspring by adding it to the population list
            results = self.population.add(offspring, use_surrogates=self.use_surrogates)
            if not self.use_surrogates:
                self._function_evaluation_count += 1

            # Replace individuals with a worse SF value than the offspring
            selected = self._select(current_neighborhood=i)

            self.population.replace(selected, offspring, results)

            # Remove the extra individual added earlier
            self.population.delete([-1])
            if self.save_non_dominated:
                self.non_dominated_archive(offspring, results)
        self._current_gen_count += 1
        self._gen_count_in_curr_iteration += 1

    def _select(self, current_neighborhood: int) -> list:
        return self.selection_operator.do(
            self.population,
            current_neighborhood,
        )



import numpy as np
from random import shuffle
#from numba import njit


class SBX_xover:
    """Simulated binary crossover.

    Parameters
        ----------
        ProC : float, optional
            [description], by default 1
        DisC : float, optional
            [description], by default 30
    """

    def __init__(self, ProC: float = 1, DisC: float = 30):
        """[summary]


        """
        self.ProC = ProC
        self.DisC = DisC

    def do(self, pop: np.ndarray, mating_pop_ids: list = None) -> np.ndarray:
        """Consecutive members of mating_pop_ids are crossed over
            in pairs. Example: if mating_pop_ids = [0, 2, 3, 6, 5] then the individuals
            are crossover as: [0, 2], [3, 6], [5, 0]. Note: if the number of elements
            is odd, the last individual is crossed over with the first one.

        Parameters
        ----------
        pop : np.ndarray
            Array of all individuals
        mating_pop_ids : list, optional
            Indices of population members to mate, by default None, which shuffles and
                mates whole population

        Returns
        -------
        np.ndarray
            The offspring produced as a result of crossover.
        """
        pop_size, num_var = pop.shape
        if mating_pop_ids is None:
            shuffled_ids = list(range(pop_size))
            shuffle(shuffled_ids)
        else:
            shuffled_ids = mating_pop_ids
        mating_pop = pop[shuffled_ids]
        mate_size = len(shuffled_ids)
        if len(shuffled_ids) % 2 == 1:
            # Maybe it should be pop_size-1?
            mating_pop = np.vstack((mating_pop, mating_pop[0]))
            mate_size = mate_size + 1
        # The rest closely follows the matlab code.
        offspring = np.zeros_like(mating_pop)  # empty_like() more efficient?
        for i in range(0, mate_size, 2):
            beta = np.zeros(num_var)
            miu = np.random.rand(num_var)
            beta[miu <= 0.5] = (2 * miu[miu <= 0.5]) ** (1 / (self.DisC + 1))
            beta[miu > 0.5] = (2 - 2 * miu[miu > 0.5]) ** (-1 / (self.DisC + 1))
            beta = beta * ((-1) ** np.random.randint(0, high=2, size=num_var))
            beta[np.random.rand(num_var) > self.ProC] = 1  # It was in matlab code
            avg = (mating_pop[i] + mating_pop[i + 1]) / 2
            diff = (mating_pop[i] - mating_pop[i + 1]) / 2
            offspring[i] = avg + beta * diff
            offspring[i + 1] = avg - beta * diff
        return offspring


# TODO: Make sure the following works correctly, then replace the code above.
"""@njit
def create_offsprings(pop:np.ndarray, mate_size:int,ProC: float=1, DisC: float=30):
    pop_size, num_var = pop.shape
    beta = np.zeros(num_var)
    offspring = np.zeros_like(pop)
    for i in range(0, mate_size, 2):
        beta[:] = 0
        miu = np.random.rand(num_var)

        beta[miu <= 0.5] = (2 * miu[miu <= 0.5]) ** (1 / (DisC + 1))
        beta[miu > 0.5] = (2 - 2 * miu[miu > 0.5]) ** (-1 / (DisC + 1))

        beta = beta * ((-1) ** np.random.randint(0, high=2, size=num_var))
        beta[np.random.rand(num_var) > ProC] = 1  # It was in matlab code

        avg = (pop[i] + pop[i + 1]) / 2
        diff = (pop[i] - pop[i + 1]) / 2
        offspring[i] = avg + beta * diff
        offspring[i + 1] = avg - beta * diff
    return offspring"""



from typing import Dict, Literal, Tuple, Type, Union

import pandas as pd
import numpy as np
from desdeo_emo.population.Population import Population
from desdeo_emo.selection.SelectionBase import SelectionBase

from desdeo_problem import MOProblem, DataProblem, EvaluationResults
from desdeo_emo.problem import IOPISProblem
from desdeo_tools.interaction import SimplePlotRequest
from desdeo_tools.interaction.request import BaseRequest
from scipy.special import comb
from desdeo_tools.utilities import non_dominated
from desdeo_emo.utilities.non_dominated import check_domination


class eaError(Exception):
    """Raised when an error related to EA occurs"""


class BaseEA:
    """This class provides the basic structure for Evolutionary algorithms."""

    def __init__(
        self,
        a_priori: bool = False,
        interact: bool = False,
        selection_operator: Type[
            SelectionBase
        ] = None,  # TODO: No algorithm uses this option. Remove?
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        use_surrogates: bool = False,
        keep_archive: bool = False,
        save_non_dominated: bool = False,
    ):
        """Initialize EA here. Set up parameters, create EA specific objects."""
        self.a_priori: bool = a_priori
        self.interact: bool = interact
        self.n_iterations: int = n_iterations
        self.n_gen_per_iter: int = n_gen_per_iter
        self.total_gen_count: int = n_gen_per_iter * n_iterations
        self.total_function_evaluations = total_function_evaluations
        self.selection_operator = selection_operator
        self.use_surrogates: bool = use_surrogates
        # Internal counters and state trackers
        self._iteration_counter: int = 0
        self._gen_count_in_curr_iteration: int = 0
        self._current_gen_count: int = 0
        self._function_evaluation_count: int = 0
        self._interaction_location: Union[
            None, Literal["Population", "Selection", "EA", "Problem"]
        ] = None
        self.interaction_type_set_bool: bool = False
        self.allowable_interaction_types: Union[None, Dict] = None
        self.population: Union[None, Population] = None
        self.keep_archive = keep_archive
        self.archive: dict = {}
        self.save_non_dominated = save_non_dominated

    """@property
    def interaction_location(
        self
    ) -> Union[None, Literal["Population", "Selection", "EA", "Problem"]]:
        ""Return the location of interaction handling
        ""
        return self._interaction_location"""

    @property
    def allowable_interaction_types(self) -> dict:
        if self._allowable_interaction_types is not None:
            return self._allowable_interaction_types
        else:
            self.allowable_interaction_types = self.set_interaction_type(None)
            return self._allowable_interaction_types

    @allowable_interaction_types.setter
    def allowable_interaction_types(self, value: dict):
        self._allowable_interaction_types = value

    def start(self):
        """Mimics the structure of the mcdm methods. Returns the request objects from self.retuests()."""
        if self.population is None:
            raise eaError("Population not initialized.")
        if self.selection_operator is None:
            raise eaError("Selection operator not initialized.")
        if self.interact and not self.interaction_type_set_bool:
            raise eaError(
                "Interaction type not set. Use the set_interaction_type() method."
            )
        if self.save_non_dominated:
            self.non_dominated = {}
            non_dom_indices = non_dominated(self.population.fitness)
            self.non_dominated["individuals"] = self.population.individuals[
                non_dom_indices
            ]
            self.non_dominated["objectives"] = self.population.objectives[
                non_dom_indices
            ]
            self.non_dominated["fitness"] = self.population.fitness[non_dom_indices]
            # TODO: add uncertainity, constraints, and generation numbers
        return self.requests()

    def end(self):
        """To be run at the end of the evolution process."""
        if self.keep_archive:
            objective_names = self.population.problem.objective_names
            if isinstance(objective_names[0], list):
                objective_names = [
                    item for sublist in objective_names for item in sublist
                ]
            fitness_names = self.population.problem.fitness_names
            if isinstance(fitness_names[0], list):
                objective_names = [
                    item for sublist in fitness_names for item in sublist
                ]
            self.archive[self._current_gen_count] = {
                "iteration": self._iteration_counter,
                "decision variables": pd.DataFrame(
                    self.population.individuals,
                    columns=self.population.problem.variable_names,
                ),
                "objectives": pd.DataFrame(
                    self.population.objectives,
                    columns=objective_names,
                ),
                "fitness": pd.DataFrame(
                    self.population.fitness,
                    columns=fitness_names,
                ),
            }
            return self.archive

    def _next_gen(self):
        """Run one generation of an EA. Change nothing about the parameters."""

    def iterate(self, preference=None) -> Tuple:
        """Run one iteration of EA.

        One iteration consists of a constant or variable number of
        generations. This method leaves EA.params unchanged, except the current
        iteration count and gen count.
        """
        self.manage_preferences(preference)
        self.pre_iteration()
        self._gen_count_in_curr_iteration = 0
        while self.continue_iteration():
            self._next_gen()
        self._iteration_counter += 1
        self.post_iteration()
        return self.requests()

    def pre_iteration(self):
        """Run this code before every iteration."""
        if self.keep_archive:
            objective_names = self.population.problem.objective_names
            if isinstance(objective_names[0], list):
                objective_names = [
                    item for sublist in objective_names for item in sublist
                ]
            fitness_names = self.population.problem.fitness_names
            if isinstance(fitness_names[0], list):
                objective_names = [
                    item for sublist in fitness_names for item in sublist
                ]
            self.archive[self._current_gen_count] = {
                "iteration": self._iteration_counter,
                "decision variables": pd.DataFrame(
                    self.population.individuals,
                    columns=self.population.problem.variable_names,
                ),
                "objectives": pd.DataFrame(
                    self.population.objectives,
                    columns=objective_names,
                ),
                "fitness": pd.DataFrame(
                    self.population.fitness,
                    columns=fitness_names,
                ),
            }

    def post_iteration(self):
        """Run this code after every iteration."""
        return

    def continue_iteration(self):
        """Checks whether the current iteration should be continued or not."""
        return (
            self._gen_count_in_curr_iteration < self.n_gen_per_iter
            and self.check_FE_count()
        )

    def continue_evolution(self) -> bool:
        """Checks whether the current iteration should be continued or not."""
        return self._iteration_counter < self.n_iterations and self.check_FE_count()

    def check_FE_count(self) -> bool:
        """Checks whether termination criteria via function evaluation count has been
            met or not.

        Returns:
            bool: True is function evaluation count limit NOT met.
        """
        if self.total_function_evaluations == 0:
            return True
        elif self._function_evaluation_count <= self.total_function_evaluations:
            return True
        return False

    def set_interaction_type(
        self, interaction_type: Union[None, str]
    ) -> Union[None, str]:
        if self._interaction_location == "Selection":
            try:
                self.interaction_type_set_bool = True
                return self.selection_operator.set_interaction_type(interaction_type)
            except NotImplementedError as e:
                self.interaction_type_set_bool = False
                raise eaError(
                    "Interaction not implemented in the selection operator."
                ) from e
        if self._interaction_location == "Problem":
            try:
                self.interaction_type_set_bool = True
                return
            except NotImplementedError as e:
                self.interaction_type_set_bool = False
                raise eaError(
                    "Interaction not implemented in the problem class."
                ) from e
        if self._interaction_location == "EA":
            self.interaction_type_set_bool = False
            raise NotImplementedError("No interaction type implemented yet.")
        if self._interaction_location == "Population":
            try:
                self.interaction_type_set_bool = True
                return self.selection_operator.set_interaction_type(interaction_type)
            except NotImplementedError as e:
                self.interaction_type_set_bool = False
                raise eaError(
                    "Interaction not implemented in the population class."
                ) from e

    def manage_preferences(self, preference=None):
        """Forward the preference to the correct preference handling method.

        Args:
            preference (_type_, optional): _description_. Defaults to None.

        Raises:
            eaError: Preference handling not implemented.
        """
        if self._interaction_location is None:
            return
        if self._interaction_location == "Population":
            try:
                self.population.manage_preferences(preference)
                return
            except NotImplementedError as e:
                raise eaError(
                    "Interaction handling not implemented in population."
                ) from e
        if self._interaction_location == "Selection":
            try:
                self.selection_operator.manage_preferences(self.population, preference)
                return
            except NotImplementedError as e:
                raise eaError(
                    "Interaction handling not implemented in selection."
                ) from e
        if self._interaction_location == "EA":
            raise eaError("Override this method in child class.")
        if self._interaction_location == "Problem":
            try:
                self.population.problem.manage_preferences(self.population, preference)
                self.population.reevaluate_fitness()
                return
            except NotImplementedError as e:
                raise eaError(
                    "Interaction handling not implemented in problem class."
                ) from e

    def requests(self) -> Tuple:
        pass

    def non_dominated_archive(
        self, individuals: np.ndarray, results: EvaluationResults
    ) -> None:
        assert self.save_non_dominated
        new_fitness = results.fitness
        new_objectives = results.objectives
        if new_fitness.ndim == 1:
            new_fitness = np.atleast_2d(new_fitness)
            new_objectives = np.atleast_2d(results.objectives)
        if individuals.ndim == 1:
            individuals = np.atleast_2d(individuals)
        new_indices, old_indices = check_domination(
            new_fitness, self.non_dominated["fitness"]
        )
        self.non_dominated["individuals"] = np.vstack(
            (self.non_dominated["individuals"][old_indices], individuals[new_indices])
        )
        self.non_dominated["objectives"] = np.vstack(
            (
                self.non_dominated["objectives"][old_indices],
                new_objectives[new_indices],
            )
        )
        self.non_dominated["fitness"] = np.vstack(
            (self.non_dominated["fitness"][old_indices], new_fitness[new_indices])
        )


class BaseDecompositionEA(BaseEA):
    """The Base class for decomposition based EAs.

    This class contains most of the code to set up the parameters and operators.
    It also contains the logic of a simple decomposition EA.

    Parameters
    ----------
    problem : MOProblem
        The problem class object specifying the details of the problem.
    selection_operator : Type[SelectionBase], optional
        The selection operator to be used by the EA, by default None.
    population_size : int, optional
        The desired population size, by default None, which sets up a default value
        of population size depending upon the dimensionaly of the problem.
    population_params : Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population : Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    lattice_resolution : int, optional
        The number of divisions along individual axes in the objective space to be
        used while creating the reference vector lattice by the simplex lattice
        design. By default None
    a_priori : bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact : bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    n_iterations : int, optional
        The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter : int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations :int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    """

    def __init__(
        self,
        problem: MOProblem = None,
        initial_population: Population = None,
        population_size: int = None,
        population_params: Dict = None,
        lattice_resolution: int = None,
        interact: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        use_surrogates: bool = False,
        total_function_evaluations: int = 0,
        keep_archive: bool = False,
        save_non_dominated: bool = False,
    ):
        super().__init__(
            interact=interact,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
            use_surrogates=use_surrogates,
            keep_archive=keep_archive,
            save_non_dominated=save_non_dominated,
        )
        if interact:
            if isinstance(problem, (MOProblem, DataProblem)):
                self._interaction_location = "Selection"
            if isinstance(problem, (IOPISProblem)):
                self._interaction_location = "Problem"
                self.set_interaction_type(None)
        if initial_population is not None:
            if problem is not None:
                raise eaError("Provide only one of initial_population or problem.")
            #  Population should be compatible.
            self.population = initial_population  # TODO put checks here.
            num_fitnesses = self.population.problem.n_of_fitnesses
            if population_size is None and lattice_resolution is None:
                population_size = self.population.pop_size
        else:
            if problem is None:
                raise eaError("Provide one of initial_population or problem.")
            num_fitnesses = problem.n_of_fitnesses
            if population_size is None:
                if lattice_resolution is None:
                    lattice_res_options = [49, 13, 7, 5, 4, 3, 3, 3, 3]
                    if num_fitnesses < 11:
                        lattice_resolution = lattice_res_options[num_fitnesses - 2]
                    else:
                        lattice_resolution = 3
                population_size = comb(
                    lattice_resolution + num_fitnesses - 1,
                    num_fitnesses - 1,
                    exact=True,
                )

            self.population = Population(
                problem, population_size, population_params, use_surrogates
            )
            self._function_evaluation_count += population_size
        # self.reference_vectors = ReferenceVectors(lattice_resolution, num_fitnesses)
        # print("Using BaseDecompositionEA init")

    def _next_gen(self):
        """Run one generation of decomposition based EA. Intended to be used by
        next_iteration.
        """
        offspring = self.population.mate()  # (params=self.params)
        if self.save_non_dominated:
            results = self.population.add(offspring, self.use_surrogates)
            self.non_dominated_archive(offspring, results)
        else:
            self.population.add(offspring, self.use_surrogates)
        selected = self._select()
        self.population.keep(selected)
        self._current_gen_count += 1
        self._gen_count_in_curr_iteration += 1
        if not self.use_surrogates:
            self._function_evaluation_count += offspring.shape[0]

    def pre_iteration(self):
        super().pre_iteration()
        if not self.interact:
            self.selection_operator.adapt_RVs(self.population.fitness)
        if self._interaction_location != "Selection":
            self.selection_operator.adapt_RVs(self.population.fitness)

    def _select(self) -> list:
        """Describe a selection mechanism. Return indices of selected
        individuals.

        Returns
        -------
        list
            List of indices of individuals to be selected.
        """
        return self.selection_operator.do(self.population)

    def request_plot(self) -> SimplePlotRequest:
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"],
            columns=self.population.problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = self.population.problem._max_multiplier
        dimensions_data.loc["ideal"] = self.population.ideal_objective_vector
        dimensions_data.loc["nadir"] = self.population.nadir_objective_vector
        data = pd.DataFrame(
            self.population.objectives, columns=self.population.problem.objective_names
        )
        return SimplePlotRequest(
            data=data, dimensions_data=dimensions_data, message="Objective Values"
        )

    def request_preferences(self) -> Type[BaseRequest]:

        if self.interact is False:
            return
        if self._interaction_location == "Problem":
            return self.population.problem.request_preferences(self.population)
        if self._interaction_location == "Selection":
            return self.selection_operator.request_preferences(self.population)

    def requests(self) -> Tuple:
        return (self.request_preferences(), self.request_plot())

    def end(self):
        """Conducts non-dominated sorting at the end of the evolution process

        Returns:
            tuple: The first element is a 2-D array of the decision vectors of the non-dominated solutions.
                The second element is a 2-D array of the corresponding objective values.
        """
        base_return = super().end()
        non_dom = self.population.non_dominated_objectives()
        return (
            self.population.individuals[non_dom, :],
            self.population.objectives[non_dom, :],
            base_return,
        )
